# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration models V9."""

import datetime
import os
import posixpath
import re
import uuid
import weakref
from bisect import bisect
from collections import OrderedDict
from copy import copy
from functools import total_ordering
from pathlib import Path
from typing import Type, Union
from urllib.parse import quote, urljoin, urlparse

import attr
from attr.validators import instance_of
from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import (
    DateTimeList,
    JsonLDSchema,
    Nested,
    StringList,
    Uri,
    fields,
    oa,
    prov,
    rdfs,
    renku,
    schema,
)
from renku.command.schema.project import ProjectSchema as V10ProjectSchema
from renku.core import errors
from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION
from renku.core.migration.models import v10 as new_schema
from renku.core.migration.models.refs import LinkReference
from renku.core.migration.utils import (
    OLD_METADATA_PATH,
    generate_dataset_file_url,
    generate_dataset_id,
    generate_dataset_tag_id,
    generate_url_id,
    get_datasets_path,
)
from renku.core.util import yaml as yaml
from renku.core.util.datetime8601 import fix_datetime, parse_date
from renku.core.util.doi import extract_doi, is_doi
from renku.core.util.git import get_in_submodules
from renku.core.util.urls import get_host, get_slug
from renku.domain_model.dataset import generate_default_slug
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Commit
from renku.version import __version__, version_url

wfprov = fields.Namespace("http://purl.org/wf4ever/wfprov#")
PROJECT_URL_PATH = "projects"
RANDOM_ID_LENGTH = 4


def _set_entity_commit(entity, commit):
    """Set the commit of an entity."""
    if not entity.commit:
        revision = "UNCOMMITTED"
        if entity._label:
            revision = entity._label.rsplit("@", maxsplit=1)[-1]
        if revision == "UNCOMMITTED":
            commit = commit
        elif project_context.has_context():
            commit = project_context.repository.get_commit(revision)
        entity.commit = commit


def _str_or_none(data):
    """Return str representation or None."""
    return str(data) if data is not None else data


def generate_project_id(name, creator):
    """Return the id for the project based on the repository origin remote."""

    # Determine the hostname for the resource URIs.
    # If RENKU_DOMAIN is set, it overrides the host from remote.
    # Default is localhost.
    host = "localhost"

    if not creator:
        raise ValueError("Project Creator not set")

    owner = creator.email.split("@")[0]

    if project_context.has_context():
        remote = project_context.remote
        host = remote.host or host
        owner = remote.owner or owner
        name = remote.name or name
    host = os.environ.get("RENKU_DOMAIN") or host
    if name:
        name = quote(name, safe="")
    else:
        raise ValueError("Project name not set")

    project_url = urljoin(f"https://{host}", posixpath.join(PROJECT_URL_PATH, owner, name))
    return project_url


@attr.s(slots=True)
class Project:
    """Represent a project."""

    name = attr.ib(default=None)

    created = attr.ib(converter=parse_date)

    version = attr.ib(converter=str, default=str(SUPPORTED_PROJECT_VERSION))

    agent_version = attr.ib(converter=str, default="pre-0.11.0")

    template_source = attr.ib(type=str, default=None)

    template_ref = attr.ib(type=str, default=None)

    template_id = attr.ib(type=str, default=None)

    template_version = attr.ib(type=str, default=None)

    template_metadata = attr.ib(type=str, default="{}")

    immutable_template_files = attr.ib(factory=list)

    automated_update = attr.ib(converter=bool, default=True)

    creator = attr.ib(default=None, kw_only=True)

    _id = attr.ib(kw_only=True, default=None)

    _metadata_path = attr.ib(default=None, init=False)

    @created.default
    def _now(self):
        """Define default value for datetime fields."""
        return datetime.datetime.now(datetime.timezone.utc)

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        if not self.creator:
            old_metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
            repository = project_context.repository
            if old_metadata_path.exists():
                self.creator = Person.from_commit(repository.get_previous_commit(old_metadata_path, first=True))
            else:
                # this assumes the project is being newly created
                self.creator = Person.from_repository(repository)

        try:
            self._id = self.project_id
        except ValueError:
            """Fallback to old behaviour."""
            if not self._id:
                try:
                    self._id = project_context.project.id
                except ValueError:
                    metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
                    self._id = Project.from_yaml(metadata_path)._id
                except errors.ConfigurationError:
                    pass
                if not self._id:
                    raise

    @property
    def project_id(self):
        """Return the id for the project."""
        return generate_project_id(name=self.name, creator=self.creator)

    @classmethod
    def from_yaml(cls, path):
        """Return an instance from a YAML file."""
        data = yaml.read_yaml(path)
        self = cls.from_jsonld(data=data)
        self._metadata_path = path

        return self

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return ProjectSchema().load(data)

    def to_yaml(self, path=None):
        """Write an instance to the referenced YAML file."""
        from renku import __version__

        self.agent_version = __version__

        self._metadata_path = path or self._metadata_path
        data = ProjectSchema().dump(self)
        yaml.write_yaml(path=self._metadata_path, data=data)


@attr.s(eq=False, order=False)
class CommitMixin:
    """Represent a commit mixin."""

    commit = attr.ib(default=None, kw_only=True)
    path = attr.ib(default=None, kw_only=True, converter=_str_or_none)

    _id = attr.ib(default=None, kw_only=True)
    _label = attr.ib(kw_only=True)
    _project = attr.ib(type=Project, kw_only=True, default=None)

    def default_id(self):
        """Configure calculated ID."""
        hexsha = self.commit.hexsha if self.commit else "UNCOMMITTED"
        return generate_file_id(hexsha=hexsha, path=self.path)

    @_label.default
    def default_label(self):
        """Generate a default label."""
        if self.commit:
            hexsha = self.commit.hexsha
        else:
            hexsha = "UNCOMMITTED"
        if self.path:
            path = self.path
            if project_context.has_context() and os.path.isabs(path):
                path = Path(path).relative_to(project_context.path)
            return generate_label(path, hexsha)
        return hexsha

    def __attrs_post_init__(self):
        """Post-init hook."""
        if self.path and project_context.has_context():
            path = Path(self.path)
            if path.is_absolute():
                self.path = str(path.relative_to(project_context.path))

        # always force "project" to be the current project
        if project_context.has_context():
            try:
                self._project = project_context.project
            except ValueError:
                metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
                self._project = Project.from_yaml(metadata_path)

        if not self._id:
            self._id = self.default_id()


@attr.s(eq=False, order=False)
class Entity(CommitMixin):
    """Represent a data value or item."""

    _parent = attr.ib(
        default=None, kw_only=True, converter=lambda value: weakref.ref(value) if value is not None else None
    )

    checksum = attr.ib(default=None, kw_only=True, type=str)

    @classmethod
    def from_revision(cls, path, revision: Union[str, Commit] = "HEAD", parent=None, find_previous=True, **kwargs):
        """Return dependency from given path and revision."""
        repository = project_context.repository

        if find_previous:
            revision = repository.get_previous_commit(path, revision=revision)
        elif revision == "HEAD":
            revision = repository.head.commit
        else:
            assert isinstance(revision, Commit)

        _, commit, path = get_in_submodules(project_context.repository, revision, path)

        path_ = project_context.path / path
        if path != "." and path_.is_dir():
            entity = Collection(commit=commit, path=path, members=[], parent=parent)

            files_in_commit = [c.b_path for c in commit.get_changes() if not c.deleted]

            # update members with commits
            for member in path_.iterdir():
                if member.name == ".gitkeep":
                    continue

                member_path = str(member.relative_to(project_context.path))
                find_previous = True

                if member_path in files_in_commit:
                    # we already know the newest commit, no need to look it up
                    find_previous = False

                try:
                    assert all(member_path != m.path for m in entity.members)

                    entity.members.append(
                        cls.from_revision(
                            path=member_path,
                            revision=commit,
                            parent=entity,
                            find_previous=find_previous,
                            **kwargs,
                        )
                    )
                except errors.GitCommitNotFoundError:
                    pass

        else:
            entity = cls(commit=commit, path=str(path), parent=parent, **kwargs)

        return entity

    @property
    def parent(self):  # pragma: no cover
        """Return the parent object."""
        return self._parent() if self._parent is not None else None

    @property
    def entities(self):
        """Yield itself."""
        if project_context.has_context() and not self.commit and self._label and "@UNCOMMITTED" not in self._label:
            repository = project_context.repository
            self.commit = repository.get_commit(self._label.rsplit("@", maxsplit=1)[-1])

        yield self


@attr.s(eq=False, order=False)
class Collection(Entity):
    """Represent a directory with files."""

    members = attr.ib(kw_only=True, default=None)

    def __attrs_post_init__(self):
        """Init members."""
        super().__attrs_post_init__()

        if self.members is None:
            self.members = self.default_members()

        for member in self.members:
            member._parent = weakref.ref(self)

    def default_members(self):
        """Generate default members as entities from current path."""
        if not project_context.has_context():
            return []
        dir_path = project_context.path / self.path

        if not dir_path.exists():
            # likely a directory deleted in a previous commit
            return []

        assert dir_path.is_dir()

        members = []
        for path in dir_path.iterdir():
            if path.name == ".gitkeep":
                continue  # ignore empty directories in Git repository
            cls: Type = Collection if path.is_dir() else Entity
            members.append(cls(commit=self.commit, path=str(path.relative_to(project_context.path)), parent=self))
        return members

    @property
    def entities(self):
        """Recursively return all files."""
        for member in self.members:
            yield from member.entities

        if project_context.has_context() and not self.commit and self._label and "@UNCOMMITTED" not in self._label:
            repository = project_context.repository
            self.commit = repository.get_commit(self._label.rsplit("@", maxsplit=1)[-1])

        yield self


@attr.s(eq=False, order=False)
class MappedIOStream:
    """Represents an IO stream (``stdin``, ``stdout``, ``stderr``)."""

    _id = attr.ib(default=None, kw_only=True)
    _label = attr.ib(default=None, kw_only=True)

    STREAMS = ["stdin", "stdout", "stderr"]

    stream_type = attr.ib(type=str, kw_only=True)

    def default_id(self):
        """Generate an id for a mapped stream."""
        host = "localhost"
        if project_context.has_context():
            host = project_context.remote.host or host
        host = os.environ.get("RENKU_DOMAIN") or host

        return urljoin(f"https://{host}", posixpath.join("/iostreams", self.stream_type))

    def default_label(self):
        """Set default label."""
        return f'Stream mapping for stream "{self.stream_type}"'

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._id:
            self._id = self.default_id()
        if not self._label:
            self._label = self.default_label()


@attr.s(eq=False, order=False)
class CommandParameter:
    """Represents a parameter for an execution template."""

    _id = attr.ib(default=None, kw_only=True)
    _label = attr.ib(default=None, kw_only=True)

    default_value = attr.ib(default=None, kw_only=True)

    description = attr.ib(default=None, kw_only=True)

    name: str = attr.ib(default=None, kw_only=True)

    position = attr.ib(default=None, type=int, kw_only=True)

    prefix = attr.ib(default=None, type=str, kw_only=True)

    @property
    def sanitized_id(self):
        """Return ``_id`` sanitized for use in non-jsonld contexts."""
        if "/steps/" in self._id:
            return "/".join(self._id.split("/")[-4:])
        return "/".join(self._id.split("/")[-2:])

    def default_label(self):
        """Set default label."""
        raise NotImplementedError

    def default_name(self):
        """Create a default name."""
        raise NotImplementedError

    def __attrs_post_init__(self):
        """Post-init hook."""
        if not self._label:
            self._label = self.default_label()
        if not self.name:
            self.name = self.default_name()


def _generate_name(base, prefix, position):
    name = get_slug(prefix.strip(" -=")) if prefix else base
    position = position or uuid.uuid4().hex[:RANDOM_ID_LENGTH]
    return f"{name}-{position}"


@attr.s(eq=False, order=False)
class CommandArgument(CommandParameter):
    """An argument to a command that is neither input nor output."""

    value = attr.ib(default=None, type=str, kw_only=True)

    @staticmethod
    def generate_id(run_id, position=None):
        """Generate an id for an argument."""
        if position:
            id_ = str(position)
        else:
            id_ = uuid.uuid4().hex
        return f"{run_id}/arguments/{id_}"

    def default_label(self):
        """Set default label."""
        return f'Command Argument "{self.default_value}"'

    def default_name(self):
        """Create a default name."""
        return _generate_name(base="param", prefix=self.prefix, position=self.position)

    def __attrs_post_init__(self):
        """Post-init hook."""
        super().__attrs_post_init__()

        if not self.default_value:
            self.default_value = self.value


@attr.s(eq=False, order=False)
class CommandInput(CommandParameter):
    """An input to a command."""

    consumes = attr.ib(kw_only=True)

    mapped_to = attr.ib(default=None, kw_only=True)

    @staticmethod
    def generate_id(run_id, position=None):
        """Generate an id for an argument."""
        if position:
            id_ = str(position)
        else:
            id_ = uuid.uuid4().hex
        return f"{run_id}/inputs/{id_}"

    def default_label(self):
        """Set default label."""
        return f'Command Input "{self.default_value}"'

    def default_name(self):
        """Create a default name."""
        return _generate_name(base="input", prefix=self.prefix, position=self.position)

    def __attrs_post_init__(self):
        """Post-init hook."""
        super().__attrs_post_init__()

        if not self.default_value:
            self.default_value = self.consumes.path


@attr.s(eq=False, order=False)
class CommandOutput(CommandParameter):
    """An output of a command."""

    create_folder = attr.ib(default=False, kw_only=True, type=bool)

    produces = attr.ib(kw_only=True)

    mapped_to = attr.ib(default=None, kw_only=True)

    @staticmethod
    def generate_id(run_id, position=None):
        """Generate an id for an argument."""
        if position:
            id_ = str(position)
        else:
            id_ = uuid.uuid4().hex
        return f"{run_id}/outputs/{id_}"

    def default_label(self):
        """Set default label."""
        return f'Command Output "{self.default_value}"'

    def default_name(self):
        """Create a default name."""
        return _generate_name(base="output", prefix=self.prefix, position=self.position)

    def __attrs_post_init__(self):
        """Post-init hook."""
        super().__attrs_post_init__()

        if not self.default_value:
            self.default_value = self.produces.path


@attr.s(eq=False, order=False)
class RunParameter:
    """A run parameter that is set inside the script."""

    _id = attr.ib(default=None, kw_only=True)

    _label = attr.ib(default=None, kw_only=True)

    name = attr.ib(default=None, type=str, kw_only=True)

    value = attr.ib(default=None, type=str, kw_only=True)

    type = attr.ib(default=None, type=str, kw_only=True)


@total_ordering
@attr.s(eq=False, order=False)
class Run(CommitMixin):
    """Represents a `renku run` execution template."""

    command = attr.ib(default=None, type=str, kw_only=True)

    successcodes = attr.ib(kw_only=True, type=list, factory=list)

    subprocesses = attr.ib(kw_only=True, factory=list)

    arguments = attr.ib(kw_only=True, factory=list)

    inputs = attr.ib(kw_only=True, factory=list)

    outputs = attr.ib(kw_only=True, factory=list)

    run_parameters = attr.ib(kw_only=True, factory=list)

    name = attr.ib(default=None, kw_only=True, type=str)

    description = attr.ib(default=None, kw_only=True, type=str)

    keywords = attr.ib(kw_only=True, factory=list)

    _activity = attr.ib(kw_only=True, default=None)

    @staticmethod
    def generate_id(identifier=None):
        """Generate an id for an argument."""
        host = "localhost"
        host = project_context.remote.host or host
        host = os.environ.get("RENKU_DOMAIN") or host

        if not identifier:
            identifier = str(uuid.uuid4())

        return urljoin(f"https://{host}", posixpath.join("/runs", quote(identifier, safe="")))

    def __lt__(self, other):
        """Compares two subprocesses order based on their dependencies."""
        a_inputs = set()
        b_outputs = set()

        for i in other.inputs:
            entity = i.consumes
            for sub_entity in entity.entities:
                a_inputs.add(sub_entity.path)

        for i in self.outputs:
            entity = i.produces
            for sub_entity in entity.entities:
                b_outputs.add(sub_entity.path)

        return a_inputs & b_outputs

    def add_subprocess(self, subprocess):
        """Adds a subprocess to this run."""
        process_order = 0
        if self.subprocesses:
            processes = [o.process for o in self.subprocesses]
            # Get position to insert based on dependencies
            process_order = bisect(processes, subprocess)
            if process_order < len(processes):
                # adjust ids of inputs inherited from latter subprocesses
                for i in range(len(processes), process_order, -1):
                    sp = self.subprocesses[i - 1]
                    sp._id = sp._id.replace(f"subprocess/{i}", f"subprocess/{i+1}")
                    sp.index += 1

                    for inp in self.inputs:
                        inp._id = inp._id.replace(f"/steps/step_{i}/", f"/steps/step_{i+1}/")
                    for outp in self.outputs:
                        outp._id = outp._id.replace(f"/steps/step_{i}/", f"/steps/step_{i+1}/")

        input_paths = [i.consumes.path for i in self.inputs]
        output_paths = [o.produces.path for o in self.outputs]

        for input_ in subprocess.inputs:
            if input_.consumes.path not in input_paths and input_.consumes.path not in output_paths:
                new_input = copy(input_)

                new_input._id = f"{self._id}/steps/step_{process_order + 1}/" f"{new_input.sanitized_id}"
                new_input.mapped_to = None

                matching_output = next((o for o in self.outputs if o.produces.path == new_input.consumes.path), None)

                if not matching_output:
                    self.inputs.append(new_input)
                    input_paths.append(new_input.consumes.path)

        for output in subprocess.outputs:
            if output.produces.path not in output_paths:
                new_output = copy(output)

                new_output._id = f"{self._id}/steps/step_{process_order + 1}/" f"{new_output.sanitized_id}"
                new_output.mapped_to = None
                self.outputs.append(new_output)
                output_paths.append(new_output.produces.path)

                matching_input = next((i for i in self.inputs if i.consumes.path == new_output.produces.path), None)
                if matching_input:
                    self.inputs.remove(matching_input)
                    input_paths.remove(matching_input.consumes.path)
        ordered_process = OrderedSubprocess(
            id=OrderedSubprocess.generate_id(self._id, process_order + 1), index=process_order + 1, process=subprocess
        )
        self.subprocesses.insert(process_order, ordered_process)


@total_ordering
@attr.s(eq=False, order=False)
class OrderedSubprocess:
    """A subprocess with ordering."""

    _id = attr.ib(kw_only=True)

    index = attr.ib(kw_only=True, type=int)

    process = attr.ib(kw_only=True)

    @staticmethod
    def generate_id(parent_id, index):
        """Generate an id for an ``OrderedSubprocess``."""
        return f"{parent_id}/subprocess/{index}"

    def __lt__(self, other):
        """Compares two ordered subprocesses."""
        return self.index < other.index


@attr.s
class Association:
    """Assign responsibility to an agent for an activity."""

    plan = attr.ib()
    agent = attr.ib(default=None)

    _id = attr.ib(kw_only=True)


class EntityProxyMixin:
    """Implement proxy to entity attribute."""

    def __getattribute__(self, name):
        """Proxy entity attributes."""
        cls = object.__getattribute__(self, "__class__")
        names = {field.name for field in attr.fields(cls)}
        names |= set(dir(cls))
        if name in names:
            return object.__getattribute__(self, name)
        entity = object.__getattribute__(self, "entity")
        return getattr(entity, name)


@attr.s(eq=False, order=False)
class Usage(EntityProxyMixin):
    """Represent a dependent path."""

    entity = attr.ib(kw_only=True)
    role = attr.ib(default=None, kw_only=True)

    _id = attr.ib(default=None, kw_only=True)


@attr.s(eq=False, order=False)
class Generation(EntityProxyMixin):
    """Represent an act of generating a file."""

    entity = attr.ib()

    role = attr.ib(default=None)

    _activity = attr.ib(
        default=None, kw_only=True, converter=lambda value: weakref.ref(value) if value is not None else None
    )
    _id = attr.ib(kw_only=True)

    @property
    def activity(self):
        """Return the activity object."""
        return self._activity() if self._activity is not None else None

    @_id.default
    def default_id(self):
        """Configure calculated ID."""
        if self.role:
            return f"{self.activity._id}/{self.role}"
        return f"{self.activity._id}/tree/{quote(str(self.entity.path))}"


@attr.s(eq=False, order=False)
class Activity(CommitMixin):
    """Represent an activity in the repository."""

    _id = attr.ib(default=None, kw_only=True)
    _message = attr.ib(kw_only=True)
    _was_informed_by = attr.ib(kw_only=True)

    part_of = attr.ib(default=None, kw_only=True)

    _collections = attr.ib(default=attr.Factory(OrderedDict), init=False, kw_only=True)
    generated = attr.ib(kw_only=True, default=None)

    invalidated = attr.ib(kw_only=True, default=None)

    influenced = attr.ib(kw_only=True)

    started_at_time = attr.ib(kw_only=True)

    ended_at_time = attr.ib(kw_only=True)

    agents = attr.ib(kw_only=True)

    _metadata_path = attr.ib(default=None, init=False)

    @classmethod
    def from_yaml(cls, path, commit=None):
        """Return an instance from a YAML file."""
        data = yaml.read_yaml(path)

        self = cls.from_jsonld(data=data, commit=commit)
        self._metadata_path = path

        return self

    @classmethod
    def from_jsonld(cls, data, commit=None):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, list):
            raise ValueError(data)

        schema = ActivitySchema

        if any(str(wfprov.WorkflowRun) in d["@type"] for d in data):
            schema = WorkflowRunSchema
        elif any(str(wfprov.ProcessRun) in d["@type"] for d in data):
            schema = ProcessRunSchema

        return schema(commit=commit, flattened=True).load(data)

    @_message.default
    def default_message(self):
        """Generate a default message."""
        if self.commit:
            return self.commit.message

    @_was_informed_by.default
    def default_was_informed_by(self):
        """List parent actions."""
        if self.commit:
            return [self.generate_id(parent) for parent in self.commit.parents]

    @started_at_time.default
    def default_started_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.authored_datetime

    @ended_at_time.default
    def default_ended_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.committed_datetime

    @agents.default
    def default_agents(self):
        """Set person agent to be the author of the commit."""
        renku_agent = SoftwareAgent(label=f"renku {__version__}", id=version_url)
        if self.commit:
            return [Person.from_commit(self.commit), renku_agent]
        return [renku_agent]

    @influenced.default
    def default_influenced(self):
        """Calculate default values."""
        return list(self._collections.values())


@attr.s(eq=False, order=False)
class ProcessRun(Activity):
    """A process run is a particular execution of a Process description."""

    __association_cls__ = Run

    generated = attr.ib(kw_only=True, default=None)

    association = attr.ib(default=None, kw_only=True)

    annotations = attr.ib(kw_only=True, default=None)

    qualified_usage = attr.ib(kw_only=True, default=None)

    run_parameter = attr.ib(kw_only=True, default=None)

    def __attrs_post_init__(self):
        """Calculate properties."""
        super().__attrs_post_init__()
        repository = project_context.repository
        commit_not_set = not self.commit or self.commit.hexsha in self._id
        if commit_not_set and Path(self.path).exists():
            self.commit = repository.get_previous_commit(self.path)

        if self.association:
            self.association.plan._activity = weakref.ref(self)
            plan = self.association.plan
            if not plan.commit:
                if self.commit:
                    plan.commit = self.commit

                if plan.inputs:
                    for i in plan.inputs:
                        _set_entity_commit(entity=i.consumes, commit=self.commit)
                if plan.outputs:
                    for o in plan.outputs:
                        _set_entity_commit(entity=o.produces, commit=self.commit)

        if self.qualified_usage and self.commit:
            usages = []
            revision = self.commit.hexsha
            for usage in self.qualified_usage:
                if not usage.commit and "@UNCOMMITTED" in usage._label:
                    usages.append(
                        Usage.from_revision(path=usage.path, role=usage.role, revision=revision, id=usage._id)
                    )
                else:
                    if not usage.commit:
                        revision = usage._label.rsplit("@", maxsplit=1)[-1]
                        usage.entity.commit = repository.get_commit(revision)

                    usages.append(usage)
            self.qualified_usage = usages

    @classmethod
    def generate_id(cls, commit_hexsha):
        """Calculate action ID."""
        host = "localhost"
        host = project_context.remote.host or host
        host = os.environ.get("RENKU_DOMAIN") or host

        return urljoin(
            f"https://{host}",
            posixpath.join("/activities", f"commit/{commit_hexsha}"),
        )

    @classmethod
    def from_run(cls, run, path, commit=None, subprocess_index=None, update_commits=False):
        """Convert a ``Run`` to a ``ProcessRun``."""
        repository = project_context.repository

        if not commit:
            commit = repository.head.commit

        usages = []

        id_ = ProcessRun.generate_id(commit.hexsha)

        if subprocess_index is not None:
            id_ = f"{id_}/steps/step_{subprocess_index}"

        for input_ in run.inputs:
            usage_id = f"{id_}/{input_.sanitized_id}"
            input_path = input_.consumes.path
            entity = input_.consumes
            if update_commits:
                commit = repository.get_previous_commit(input_path, revision=commit.hexsha)
                entity = Entity.from_revision(input_path, commit)

            dependency = Usage(entity=entity, role=input_.sanitized_id, id=usage_id)

            usages.append(dependency)

        agent = SoftwareAgent.from_commit(commit)
        association = Association(agent=agent, id=id_ + "/association", plan=run)

        run_parameter = []

        for parameter in run.run_parameters:
            parameter_id = f"{id_}/{parameter.name}"
            run_parameter.append(RunParameter(name=parameter.name, value=parameter.value, id=parameter_id))

        process_run = cls(
            id=id_,
            qualified_usage=usages,
            association=association,
            commit=commit,
            path=path,
            run_parameter=run_parameter,
        )

        generated = []

        for output in run.outputs:
            entity = Entity.from_revision(output.produces.path, revision=commit, parent=output.produces.parent)

            generation = Generation(activity=process_run, role=output.sanitized_id, entity=entity)
            generated.append(generation)

        process_run.generated = generated

        return process_run

    def to_yaml(self, path=None):
        """Write an instance to the referenced YAML file."""
        self._metadata_path = path or self._metadata_path
        data = ProcessRunSchema(flattened=True).dump(self)
        yaml.write_yaml(path=self._metadata_path, data=data)


@attr.s(eq=False, order=False)
class WorkflowRun(ProcessRun):
    """A workflow run typically contains several subprocesses."""

    __association_cls__ = Run

    _processes = attr.ib(kw_only=True, default=attr.Factory(list))

    @property
    def subprocesses(self):
        """Subprocesses of this ``WorkflowRun``."""
        return {i: p for i, p in enumerate(self._processes)}


@attr.s
class Url:
    """Represents a schema URL reference."""

    url = attr.ib(default=None, kw_only=True)

    url_str = attr.ib(default=None, kw_only=True)
    url_id = attr.ib(default=None, kw_only=True)

    _id = attr.ib(default=None, kw_only=True)

    def default_id(self):
        """Define default value for id field."""
        return generate_url_id(url_str=self.url_str, url_id=self.url_id)

    def default_url(self):
        """Define default value for url field."""
        if self.url_str:
            return self.url_str
        elif self.url_id:
            return {"@id": self.url_id}
        else:
            raise NotImplementedError("Either url_id or url_str has to be set")

    @property
    def value(self):
        """Returns the url value as string."""
        if self.url_str:
            return self.url_str
        elif self.url_id:
            return self.url_id
        else:
            raise NotImplementedError("Either url_id or url_str has to be set")

    def __attrs_post_init__(self):
        """Post-initialize attributes."""
        if not self.url:
            self.url = self.default_url()
        elif isinstance(self.url, dict):
            if "_id" in self.url:
                self.url["@id"] = self.url.pop("_id")
            self.url_id = self.url["@id"]
        elif isinstance(self.url, str):
            self.url_str = self.url

        if not self._id or self._id.startswith("_:"):
            self._id = self.default_id()

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return OldUrlSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return OldUrlSchema().dump(self)


def _convert_creators(value):
    """Convert creators."""
    if isinstance(value, dict):  # compatibility with previous versions
        return [Person.from_jsonld(value)]

    if isinstance(value, list):
        return [Person.from_jsonld(v) for v in value]

    return value


class Person:
    """Represent a person."""

    __slots__ = ("affiliation", "alternate_name", "email", "id", "label", "name")

    def __init__(
        self,
        *,
        affiliation: str = None,
        alternate_name: str = None,
        email: str = None,
        id: str = None,
        label: str = None,
        name: str,
    ):
        self.validate_email(email)

        if id == "mailto:None" or not id or id.startswith("_:"):
            full_identity = Person.get_full_identity(email, affiliation, name)
            id = Person.generate_id(email, full_identity, hostname=get_host(use_project_context=False))
        label = label or name

        self.affiliation: str = affiliation
        self.alternate_name: str = alternate_name
        self.email: str = email
        self.id: str = id
        self.label: str = label
        self.name: str = name

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Person):
            return False
        return self.id == other.id and self.full_identity == other.full_identity

    def __hash__(self):
        return hash((self.id, self.full_identity))

    @staticmethod
    def generate_id(email, full_identity, hostname):
        """Generate identifier for Person."""
        if email:
            return f"mailto:{email}"

        id = full_identity or str(uuid.uuid4().hex)
        id = quote(id, safe="")

        # TODO: Remove hostname part once migrating to new metadata
        return f"https://{hostname}/persons/{id}"

    @staticmethod
    def validate_email(email):
        """Check that the email is valid."""
        if not email:
            return
        if not isinstance(email, str) or not re.match(r"[^@]+@[^@]+\.[^@]+", email):
            raise ValueError("Email address is invalid.")

    @classmethod
    def from_commit(cls, commit):
        """Create an instance from a Git commit."""
        return cls(name=commit.author.name, email=commit.author.email)

    @property
    def short_name(self):
        """Gives full name in short form."""
        names = self.name.split()
        if len(names) == 1:
            return self.name

        last_name = names[-1]
        initials = [name[0] for name in names]
        initials.pop()

        return "{}.{}".format(".".join(initials), last_name)

    @property
    def full_identity(self):
        """Return name, email, and affiliation."""
        return self.get_full_identity(self.email, self.affiliation, self.name)

    @staticmethod
    def get_full_identity(email, affiliation, name):
        """Return name, email, and affiliation."""
        email = f" <{email}>" if email else ""
        affiliation = f" [{affiliation}]" if affiliation else ""
        return f"{name}{email}{affiliation}"

    @classmethod
    def from_repository(cls, repository):
        """Create an instance from a repository."""
        user = repository.get_user()
        return cls(email=user.email, name=user.name)

    @classmethod
    def from_string(cls, string):
        """Create an instance from a 'Name <email>' string."""
        regex_pattern = r"([^<>\[\]]*)" r"(?:<{1}\s*(\S+@\S+\.\S+){0,1}\s*>{1}){0,1}\s*" r"(?:\[{1}(.*)\]{1}){0,1}"
        name, email, affiliation = re.search(regex_pattern, string).groups()
        if name:
            name = name.strip()
        if affiliation:
            affiliation = affiliation.strip()
        affiliation = affiliation or None

        return cls(affiliation=affiliation, email=email, name=name)

    @classmethod
    def from_dict(cls, data):
        """Create and instance from a dictionary."""
        return cls(**data)

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return OldPersonSchema().load(data)


class SoftwareAgent:
    """Represent executed software."""

    __slots__ = ("id", "label")

    def __init__(self, *, id: str, label: str):
        self.id: str = id
        self.label: str = label

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, SoftwareAgent):
            return False
        return self.id == other.id and self.label == other.label

    def __hash__(self):
        return hash((self.id, self.label))

    @classmethod
    def from_commit(cls, commit):
        """Create an instance from a Git commit."""
        # FIXME: This method can return a Person object but SoftwareAgent is not its super class
        author = Person.from_commit(commit)
        if commit.author != commit.committer:
            return cls(label=commit.committer.name, id=commit.committer.email)
        return author


@attr.s
class CreatorMixin:
    """Mixin for handling creators container."""

    creators = attr.ib(kw_only=True, converter=_convert_creators)

    @property
    def creators_csv(self):
        """Comma-separated list of creators associated with dataset."""
        return ", ".join(creator.name for creator in self.creators)

    @property
    def creators_full_csv(self):
        """Comma-separated list of creators with full identity."""
        return ", ".join(creator.full_identity for creator in self.creators)


def _extract_doi(value):
    """Return either a string or the doi part of a URL."""
    value = str(value)
    if is_doi(value):
        return extract_doi(value)
    return value


@attr.s(slots=True)
class DatasetTag:
    """Represents a Tag of an instance of a dataset."""

    name = attr.ib(default=None, kw_only=True, validator=instance_of(str))

    description = attr.ib(default=None, kw_only=True, validator=instance_of(str))

    commit = attr.ib(default=None, kw_only=True, validator=instance_of(str))

    created = attr.ib(converter=parse_date, kw_only=True)

    dataset = attr.ib(default=None, kw_only=True)

    _id = attr.ib(default=None, kw_only=True)

    @created.default
    def _now(self):
        """Define default value for datetime fields."""
        return datetime.datetime.now(datetime.timezone.utc)

    def default_id(self):
        """Define default value for id field."""
        return generate_dataset_tag_id(name=self.name, commit=self.commit)

    def __attrs_post_init__(self):
        """Post-Init hook."""
        if not self._id or self._id.startswith("_:"):
            self._id = self.default_id()

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return OldDatasetTagSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return OldDatasetTagSchema().dump(self)


@attr.s(slots=True)
class Language:
    """Represent a language of an object."""

    alternate_name = attr.ib(default=None, kw_only=True)
    name = attr.ib(default=None, kw_only=True)

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return OldLanguageSchema().load(data)


def convert_filename_path(p):
    """Return name of the file."""
    if p:
        return Path(p).name


def convert_based_on(v):
    """Convert based_on to DatasetFile."""
    if v:
        return DatasetFile.from_jsonld(v)


@attr.s(slots=True)
class DatasetFile(Entity):
    """Represent a file in a dataset."""

    added = attr.ib(converter=parse_date, kw_only=True)

    checksum = attr.ib(default=None, kw_only=True)

    filename = attr.ib(kw_only=True, converter=convert_filename_path)

    name = attr.ib(kw_only=True, default=None)

    filesize = attr.ib(default=None, kw_only=True)

    filetype = attr.ib(default=None, kw_only=True)

    url = attr.ib(default=None, kw_only=True)

    based_on = attr.ib(default=None, kw_only=True, converter=convert_based_on)

    external = attr.ib(default=False, kw_only=True)

    source = attr.ib(default=None, kw_only=True)

    @added.default
    def _now(self):
        """Define default value for datetime fields."""
        return datetime.datetime.now(datetime.timezone.utc)

    @filename.default
    def default_filename(self):
        """Generate default filename based on path."""
        if self.path:
            return Path(self.path).name

    def default_url(self):
        """Generate default url based on project's ID."""
        return generate_dataset_file_url(filepath=self.path)

    @property
    def commit_sha(self):
        """Return commit hash."""
        return self.commit.hexsha if self.commit else ""

    @property
    def full_path(self):
        """Return full path in the current reference frame."""
        path = project_context.path / self.path
        return Path(os.path.abspath(path))

    def __attrs_post_init__(self):
        """Set the property "name" after initialization."""
        super().__attrs_post_init__()

        if not self.filename:
            self.filename = self.default_filename()

        if not self.name:
            self.name = self.filename

        parsed_id = urlparse(self._id)

        if not parsed_id.scheme:
            self._id = f"file://{self._id}"

        if not self.url:
            self.url = self.default_url()

    def update_commit(self, commit):
        """Set commit and update associated fields."""
        self.commit = commit
        self._id = self.default_id()
        self._label = self.default_label()

    def update_metadata(self, path, commit):
        """Update files metadata."""
        self.path = str((project_context.path / path).relative_to(project_context.path))
        self.update_commit(commit)
        self.filename = self.default_filename()
        self.url = self.default_url()
        self.added = self._now()

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return OldDatasetFileSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return OldDatasetFileSchema().dump(self)


def _convert_dataset_files(value):
    """Convert dataset files."""
    coll = value

    if isinstance(coll, dict):  # compatibility with previous versions
        if any([key.startswith("@") for key in coll.keys()]):
            return [DatasetFile.from_jsonld(coll)]
        else:
            coll = value.values()

    return [DatasetFile.from_jsonld(v) for v in coll]


def _convert_dataset_tags(value):
    """Convert dataset tags."""
    if isinstance(value, dict):  # compatibility with previous versions
        value = [value]

    return [DatasetTag.from_jsonld(v) for v in value]


def _convert_language(obj):
    """Convert language object."""
    return Language.from_jsonld(obj) if isinstance(obj, dict) else obj


def _convert_keyword(keywords):
    """Convert keywords collection."""
    if isinstance(keywords, (list, tuple)):
        return keywords

    if isinstance(keywords, dict):
        return keywords.keys()


@attr.s
class Dataset(Entity, CreatorMixin):
    """Represent a dataset."""

    _id = attr.ib(default=None, kw_only=True)
    _label = attr.ib(default=None, kw_only=True)

    date_published = attr.ib(default=None, kw_only=True)

    description = attr.ib(default=None, kw_only=True)

    identifier = attr.ib(default=attr.Factory(uuid.uuid4), kw_only=True, converter=_extract_doi)

    in_language = attr.ib(default=None, converter=_convert_language, kw_only=True)

    images = attr.ib(default=None, kw_only=True)

    keywords = attr.ib(converter=_convert_keyword, kw_only=True, default=None)

    license = attr.ib(default=None, kw_only=True)

    title = attr.ib(default=None, type=str, kw_only=True)

    url = attr.ib(default=None, kw_only=True)

    version = attr.ib(default=None, kw_only=True)

    date_created = attr.ib(converter=parse_date, kw_only=True)

    files = attr.ib(factory=list, converter=_convert_dataset_files, kw_only=True)

    tags = attr.ib(factory=list, converter=_convert_dataset_tags, kw_only=True)

    same_as = attr.ib(default=None, kw_only=True)

    name = attr.ib(default=None, kw_only=True)

    derived_from = attr.ib(default=None, kw_only=True)

    immutable = attr.ib(default=False, kw_only=True)

    _modified = attr.ib(default=False, init=False)

    _mutated = attr.ib(default=False, init=False)

    _metadata_path = attr.ib(default=None, init=False)

    @date_created.default
    def _now(self):
        """Define default value for datetime fields."""
        return datetime.datetime.now(datetime.timezone.utc)

    @property
    def short_id(self):
        """Shorter version of identifier."""
        if is_doi(self.identifier):
            return self.identifier
        return str(self.identifier)[:8]

    @property
    def creators_csv(self):
        """Comma-separated list of creators associated with dataset."""
        return ", ".join(creator.name for creator in self.creators)

    @property
    def keywords_csv(self):
        """Comma-separated list of keywords associated with dataset."""
        return ", ".join(self.keywords or [])

    @property
    def tags_csv(self):
        """Comma-separated list of tags associated with dataset."""
        return ",".join(tag.name for tag in self.tags)

    @property
    def initial_identifier(self):
        """Return the first identifier of the dataset."""
        if self.path:
            return Path(self.path).name

    def contains_any(self, files):
        """Check if files are already within a dataset."""
        for file_ in files:
            if self.find_file(file_["path"]):
                return True
        return False

    def find_files(self, paths):
        """Return all paths that are in files container."""
        files_paths = {str(project_context.path / f.path) for f in self.files}
        return {p for p in paths if str(p) in files_paths}

    def find_file(self, path, return_index=False):
        """Find a file in files container using its relative path."""
        for index, file in enumerate(self.files):
            if str(file.path) == str(path):
                if return_index:
                    return index
                return file

    def update_metadata(self, **kwargs):
        """Updates instance attributes."""
        for attribute, value in kwargs.items():
            if value and value != getattr(self, attribute):
                self._modified = True
                setattr(self, attribute, value)

        return self

    def update_files(self, files):
        """Update files with collection of DatasetFile objects."""
        if isinstance(files, DatasetFile):
            files = (files,)

        new_files = []

        for new_file in files:
            old_file = self.find_file(new_file.path)
            if not old_file:
                new_files.append(new_file)
            elif new_file.commit != old_file.commit or new_file.added != old_file.added:
                self.unlink_file(new_file.path)
                new_files.append(new_file)

        if not new_files:
            return

        self._modified = True
        self.files += new_files

    def unlink_file(self, path, missing_ok=False):  # FIXME: Remove unused code
        """Unlink a file from dataset.

        Args:
            path: Relative path used as key inside files container.
            missing_ok (bool): Whether to ignore missing files or raise an error (Default value = False).
        """
        index = self.find_file(path, return_index=True)
        if index is not None:
            self._modified = True
            return self.files.pop(index)

        if not missing_ok:
            raise errors.InvalidFileOperation(f"File cannot be found: {path}")

    def mutate(self):
        """Update mutation history and assign a new identifier.

        Do not mutate more than once before committing the metadata or otherwise there would be missing links in the
        chain of changes.
        """
        if self.immutable:
            raise errors.OperationError(f"Cannot mutate an immutable dataset: {self.name}")

        # As a safetynet, we only allow one mutation during lifetime of a dataset object; this is not 100% error-proof
        # because one can create a new object from a mutated but uncommitted metadata file.
        if self._mutated:
            return
        self._mutated = True

        self.same_as = None
        self.derived_from = Url(url_id=self._id)

        repository = project_context.repository
        mutator = Person.from_repository(repository)
        if not any(c for c in self.creators if c.email == mutator.email):
            self.creators.append(mutator)

        self.date_created = self._now()
        self.date_published = None

        self._replace_identifier(new_identifier=str(uuid.uuid4()))

    def _replace_identifier(self, new_identifier):
        """Replace identifier and update all related fields."""
        self.identifier = new_identifier
        self._set_id()
        self.url = self._id
        self._label = self.identifier

    def _set_id(self):
        self._id = generate_dataset_id(identifier=self.identifier)

    def __attrs_post_init__(self):
        """Post-Init hook."""
        super().__attrs_post_init__()

        self._set_id()
        self.url = self._id
        self._label = self.identifier

        if self.derived_from:
            host = get_host()
            derived_from_id = self.derived_from._id
            derived_from_url = self.derived_from.url.get("@id")
            u = urlparse(derived_from_url)
            derived_from_url = u._replace(netloc=host).geturl()
            self.derived_from = Url(id=derived_from_id, url_id=derived_from_url)

        # if `date_published` is set, we are probably dealing with
        # an imported dataset so `date_created` is not needed
        if self.date_published:
            self.date_created = None

        if not self.path:
            absolute_path = LinkReference(
                metadata_path=project_context.metadata_path, name=f"datasets/{self.name}"
            ).reference.parent
            self.path = str(absolute_path.relative_to(project_context.path))

        if project_context.has_context():
            try:
                revision = self.commit.hexsha if self.commit else "HEAD"
                repository = project_context.repository
                self.commit = repository.get_previous_commit(os.path.join(self.path, "metadata.yml"), revision=revision)
            except errors.GitCommitNotFoundError:
                pass

        if not self.name:
            self.name = generate_default_slug(self.title, self.version)

    @classmethod
    def from_yaml(cls, path, commit=None):
        """Return an instance from a YAML file."""
        data = yaml.read_yaml(path)

        self = cls.from_jsonld(data=data, commit=commit)
        self._metadata_path = path

        return self

    @classmethod
    def from_jsonld(cls, data, commit=None, schema_class=None):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, (dict, list)):
            raise ValueError(data)

        schema_class = schema_class or OldDatasetSchema
        return schema_class(commit=commit, flattened=True).load(data)

    def to_yaml(self, path=None, immutable=False):
        """Write an instance to the referenced YAML file."""
        if self._modified and not (immutable or self.immutable):
            self.mutate()

        self._metadata_path = path or self._metadata_path
        data = OldDatasetSchema(flattened=True).dump(self)
        yaml.write_yaml(path=self._metadata_path, data=data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return OldDatasetSchema(flattened=True).dump(self)


class ImageObject:
    """Represents a schema.org `ImageObject`."""

    def __init__(self, content_url: str, position: int, id=None):
        self.content_url = content_url
        self.position = position
        self.id = id

    @staticmethod
    def generate_id(dataset: Dataset, position: int) -> str:
        """Generate @id field."""
        return urljoin(dataset._id + "/", posixpath.join("images", str(position)))

    @property
    def is_absolute(self):
        """Whether content_url is an absolute or relative url."""
        return bool(urlparse(self.content_url).netloc)


class OldPersonSchema(JsonLDSchema):
    """Person schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Person, schema.Person]
        model = Person
        unknown = EXCLUDE

    affiliation = StringList(schema.affiliation, load_default=None)
    alternate_name = StringList(schema.alternateName, load_default=None)
    email = fields.String(schema.email, load_default=None)
    id = fields.Id()
    label = StringList(rdfs.label, load_default=None)
    name = StringList(schema.name, load_default=None)


class ProjectSchema(JsonLDSchema):
    """Project Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [schema.Project, prov.Location]
        model = Project
        unknown = EXCLUDE

    name = fields.String(schema.name, load_default=None)
    created = DateTimeList(schema.dateCreated, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    version = StringList(schema.schemaVersion, load_default="1")
    agent_version = StringList(schema.agent, load_default="pre-0.11.0")
    template_source = fields.String(renku.templateSource, load_default=None)
    template_ref = fields.String(renku.templateReference, load_default=None)
    template_id = fields.String(renku.templateId, load_default=None)
    template_version = fields.String(renku.templateVersion, load_default=None)
    template_metadata = fields.String(renku.templateMetadata, load_default=None)
    immutable_template_files = fields.List(renku.immutableTemplateFiles, fields.String(), load_default=[])
    automated_update = fields.Boolean(renku.automatedTemplateUpdate, load_default=True)
    creator = Nested(schema.creator, OldPersonSchema, load_default=None)
    _id = fields.Id(init_name="id", load_default=None)

    @pre_dump
    def fix_datetimes(self, obj, many=False, **kwargs):
        """Pre dump hook."""
        if many:
            return [self.fix_datetimes(o, many=False, **kwargs) for o in obj]
        obj.created = fix_datetime(obj.created)
        return obj


class OldCommitMixinSchema(JsonLDSchema):
    """CommitMixin schema."""

    class Meta:
        """Meta class."""

        model = CommitMixin

    path = fields.String(prov.atLocation)
    _id = fields.Id(init_name="id")
    _label = fields.String(rdfs.label, init_name="label", load_default=None)
    _project = Nested(
        schema.isPartOf, [ProjectSchema, "V9ProjectSchema", V10ProjectSchema], init_name="project", load_default=None
    )


class OldEntitySchema(OldCommitMixinSchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity, wfprov.Artifact]
        model = Entity

    checksum = fields.String(renku.checksum, load_default=None)


class OldCollectionSchema(OldEntitySchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Collection]
        model = Collection

    members = Nested(prov.hadMember, [OldEntitySchema, "OldCollectionSchema"], many=True)


class OldSoftwareAgentSchema(JsonLDSchema):
    """SoftwareAgent schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.SoftwareAgent, wfprov.WorkflowEngine]
        model = SoftwareAgent
        unknown = EXCLUDE

    label = fields.String(rdfs.label)
    id = fields.Id()


class OldCreatorMixinSchema(JsonLDSchema):
    """CreatorMixin schema."""

    class Meta:
        """Meta class."""

        unknown = EXCLUDE

    creators = Nested(schema.creator, OldPersonSchema, many=True)


class OldUrlSchema(JsonLDSchema):
    """Url schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.URL
        model = Url
        unknown = EXCLUDE

    url = Uri(schema.url, load_default=None)
    _id = fields.Id(init_name="id", load_default=None)


class OldDatasetTagSchema(JsonLDSchema):
    """DatasetTag schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.PublicationEvent
        model = DatasetTag
        unknown = EXCLUDE

    name = fields.String(schema.name)
    description = fields.String(schema.description, load_default=None)
    commit = fields.String(schema.location)
    created = fields.DateTime(schema.startDate, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    dataset = fields.String(schema.about)
    _id = fields.Id(init_name="id")

    @pre_dump
    def fix_datetimes(self, obj, many=False, **kwargs):
        """Pre dump hook."""
        if many:
            return [self.fix_datetimes(o, many=False, **kwargs) for o in obj]
        object.__setattr__(obj, "created", fix_datetime(obj.created))
        return obj


class OldLanguageSchema(JsonLDSchema):
    """Language schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Language
        model = Language
        unknown = EXCLUDE

    alternate_name = fields.String(schema.alternateName)
    name = fields.String(schema.name)


class OldDatasetFileSchema(OldEntitySchema):
    """DatasetFile schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.DigitalDocument
        model = DatasetFile
        unknown = EXCLUDE

    added = DateTimeList(schema.dateCreated, format="iso", extra_formats=("%Y-%m-%d",))
    name = fields.String(schema.name, load_default=None)
    url = fields.String(schema.url, load_default=None)
    based_on = Nested(schema.isBasedOn, "OldDatasetFileSchema", load_default=None)
    external = fields.Boolean(renku.external, load_default=False)
    source = fields.String(renku.source, load_default=None)

    @pre_dump
    def fix_datetimes(self, obj, many=False, **kwargs):
        """Pre dump hook."""
        if many:
            return [self.fix_datetimes(o, many=False, **kwargs) for o in obj]
        obj.added = fix_datetime(obj.added)
        return obj


class OldImageObjectSchema(JsonLDSchema):
    """ImageObject schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.ImageObject
        model = ImageObject
        unknown = EXCLUDE

    id = fields.Id(load_default=None)
    content_url = fields.String(schema.contentUrl)
    position = fields.Integer(schema.position)


class OldDatasetSchema(OldEntitySchema, OldCreatorMixinSchema):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Dataset
        model = Dataset
        unknown = EXCLUDE

    _id = fields.Id(init_name="id", load_default=None)
    _label = fields.String(rdfs.label, init_name="label", load_default=None)
    date_published = fields.DateTime(
        schema.datePublished,
        load_default=None,
        allow_none=True,
        format="%Y-%m-%d",
        extra_formats=("iso", "%Y-%m-%dT%H:%M:%S"),
    )
    description = fields.String(schema.description, load_default=None)
    identifier = fields.String(schema.identifier)
    in_language = Nested(schema.inLanguage, OldLanguageSchema, load_default=None)
    images = fields.Nested(schema.image, OldImageObjectSchema, many=True, load_default=None, allow_none=True)
    keywords = fields.List(schema.keywords, fields.String(), allow_none=True, load_default=None)
    license = Uri(schema.license, allow_none=True, load_default=None)
    title = fields.String(schema.name)
    url = fields.String(schema.url, load_default=None)
    version = fields.String(schema.version, load_default=None)
    date_created = fields.DateTime(
        schema.dateCreated, load_default=None, allow_none=True, format="iso", extra_formats=("%Y-%m-%d",)
    )
    files = Nested(schema.hasPart, OldDatasetFileSchema, many=True)
    tags = Nested(schema.subjectOf, OldDatasetTagSchema, many=True)
    same_as = Nested(schema.sameAs, OldUrlSchema, load_default=None)
    name = fields.String(schema.alternateName)
    derived_from = Nested(prov.wasDerivedFrom, OldUrlSchema, load_default=None)

    @pre_dump
    def fix_datetimes(self, obj, many=False, **kwargs):
        """Pre dump hook."""
        if many:
            return [self.fix_datetimes(o, many=False, **kwargs) for o in obj]
        obj.date_published = fix_datetime(obj.date_published)
        obj.date_created = fix_datetime(obj.date_created)
        return obj


def get_project_datasets():
    """Return Dataset migration models for a project."""
    paths = get_datasets_path().rglob(OLD_METADATA_PATH)
    return [Dataset.from_yaml(path=path) for path in paths]


def generate_label(path, hexsha):
    """Generate label field."""
    return f"{path}@{hexsha}"


def generate_file_id(hexsha, path):
    """Generate DatasetFile id field."""
    # Determine the hostname for the resource URIs.
    # If RENKU_DOMAIN is set, it overrides the host from remote.
    # Default is localhost.
    host = "localhost"
    host = project_context.remote.host or host
    host = os.environ.get("RENKU_DOMAIN") or host

    # TODO: Use plural name for entity id: /blob/ -> /blobs/
    # always set the id by the identifier
    return urljoin(f"https://{host}", posixpath.join(f"/blob/{hexsha}/{quote(str(path))}"))


class MappedIOStreamSchema(JsonLDSchema):
    """MappedIOStream schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.IOStream]
        model = MappedIOStream
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    _label = fields.String(rdfs.label, init_name="label")
    stream_type = fields.String(renku.streamType)


class CommandParameterSchema(JsonLDSchema):
    """CommandParameter schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandParameter]  # , schema.PropertyValueSpecification]
        model = CommandParameter
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    _label = fields.String(rdfs.label, init_name="label")
    default_value = fields.Raw(schema.defaultValue, load_default=None)
    description = fields.String(schema.description, load_default=None)
    name = fields.String(schema.name, load_default=None)
    position = fields.Integer(renku.position, load_default=None)
    prefix = fields.String(renku.prefix, load_default=None)


class CommandArgumentSchema(CommandParameterSchema):
    """CommandArgument schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandArgument]
        model = CommandArgument
        unknown = EXCLUDE

    value = fields.String(renku.value)


class CommandInputSchema(CommandParameterSchema):
    """CommandArgument schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandInput]
        model = CommandInput
        unknown = EXCLUDE

    consumes = Nested(renku.consumes, [OldEntitySchema, OldCollectionSchema])
    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, load_default=None)


class CommandOutputSchema(CommandParameterSchema):
    """CommandArgument schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandOutput]
        model = CommandOutput
        unknown = EXCLUDE

    create_folder = fields.Boolean(renku.createFolder)
    produces = Nested(renku.produces, [OldEntitySchema, OldCollectionSchema])
    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, load_default=None)


class RunParameterSchema(JsonLDSchema):
    """RunParameter schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.RunParameter]
        model = RunParameter
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    _label = fields.String(rdfs.label, init_name="label")
    name = fields.String(schema.name)
    value = fields.String(renku.value)
    type = fields.String(renku.type)


class RunSchema(OldCommitMixinSchema):
    """Run schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.Run, prov.Plan, prov.Entity]
        model = Run
        unknown = EXCLUDE

    command = fields.String(renku.command, load_default=None)
    successcodes = fields.List(renku.successCodes, fields.Integer(), load_default=[0])
    subprocesses = Nested(renku.hasSubprocess, nested="OrderedSubprocessSchema", load_default=None, many=True)
    arguments = Nested(renku.hasArguments, CommandArgumentSchema, many=True, load_default=None)
    inputs = Nested(renku.hasInputs, CommandInputSchema, many=True, load_default=None)
    outputs = Nested(renku.hasOutputs, CommandOutputSchema, many=True, load_default=None)
    run_parameters = Nested(renku.hasRunParameters, RunParameterSchema, many=True, load_default=None)
    name = fields.String(schema.name, load_default=None)
    description = fields.String(schema.description, load_default=None)
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)


class OrderedSubprocessSchema(JsonLDSchema):
    """OrderedSubprocess schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.OrderedSubprocess]
        model = OrderedSubprocess
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    index = fields.Integer(renku.index)
    process = Nested(renku.process, RunSchema)


class AssociationSchema(JsonLDSchema):
    """Association schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Association
        model = Association
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    plan = Nested(prov.hadPlan, [RunSchema])
    agent = Nested(prov.agent, [OldSoftwareAgentSchema, OldPersonSchema])


class UsageSchema(JsonLDSchema):
    """Usage schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Usage
        model = Usage
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    entity = Nested(prov.entity, [OldEntitySchema, OldCollectionSchema, OldDatasetSchema, OldDatasetFileSchema])
    role = fields.String(prov.hadRole, load_default=None)


class GenerationSchema(JsonLDSchema):
    """Generation schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Generation
        model = Generation
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    entity = Nested(
        prov.qualifiedGeneration,
        [OldEntitySchema, OldCollectionSchema, OldDatasetSchema, OldDatasetFileSchema],
        reverse=True,
    )
    role = fields.String(prov.hadRole, load_default=None)


class ActivitySchema(OldCommitMixinSchema):
    """Activity schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Activity
        model = Activity
        unknown = EXCLUDE

    _message = fields.String(rdfs.comment, init_name="message", load_default=None)
    _was_informed_by = fields.List(prov.wasInformedBy, fields.IRI(), init_name="was_informed_by")
    generated = Nested(prov.activity, GenerationSchema, reverse=True, many=True, load_default=None)
    invalidated = Nested(
        prov.wasInvalidatedBy, [OldEntitySchema, OldCollectionSchema], reverse=True, many=True, load_default=None
    )
    influenced = Nested(prov.influenced, OldCollectionSchema, many=True)
    started_at_time = fields.DateTime(prov.startedAtTime, format="iso", add_value_types=True)
    ended_at_time = fields.DateTime(prov.endedAtTime, format="iso", add_value_types=True)
    agents = Nested(prov.wasAssociatedWith, [OldPersonSchema, OldSoftwareAgentSchema], many=True)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.started_at_time = obj.started_at_time.replace(microsecond=0)
            obj.ended_at_time = obj.ended_at_time.replace(microsecond=0)

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class ProcessRunSchema(ActivitySchema):
    """ProcessRun schema."""

    class Meta:
        """Meta class."""

        rdf_type = wfprov.ProcessRun
        model = ProcessRun
        unknown = EXCLUDE

    association = Nested(prov.qualifiedAssociation, AssociationSchema)
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    qualified_usage = Nested(prov.qualifiedUsage, UsageSchema, many=True)
    run_parameter = Nested(renku.hasRunParameter, RunParameterSchema, many=True)


class WorkflowRunSchema(ProcessRunSchema):
    """WorkflowRun schema."""

    class Meta:
        """Meta class."""

        rdf_type = wfprov.WorkflowRun
        model = WorkflowRun
        unknown = EXCLUDE

    _processes = Nested(wfprov.wasPartOfWorkflowRun, ProcessRunSchema, reverse=True, many=True, init_name="processes")


class V9ProjectSchema(JsonLDSchema):
    """Project Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [schema.Project, prov.Location]
        model = new_schema.Project
        unknown = EXCLUDE

    agent_version = StringList(schema.agent, load_default="pre-0.11.0")
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    automated_update = fields.Boolean(renku.automatedTemplateUpdate, load_default=True)
    creator = Nested(schema.creator, PersonSchema, load_default=None)
    date_created = DateTimeList(schema.dateCreated, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    description = fields.String(schema.description, load_default=None)
    id = fields.Id(load_default=None)
    immutable_template_files = fields.List(renku.immutableTemplateFiles, fields.String(), load_default=list())
    name = fields.String(schema.name, load_default=None)
    template_id = fields.String(renku.templateId, load_default=None)
    template_metadata = fields.String(renku.templateMetadata, load_default=None)
    template_ref = fields.String(renku.templateReference, load_default=None)
    template_source = fields.String(renku.templateSource, load_default=None)
    template_version = fields.String(renku.templateVersion, load_default=None)
    version = StringList(schema.schemaVersion, load_default="1")
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku migrations management.

Migrations files are put in renku/core/management/migrations directory. Name
of these files has ``m_1234__name.py`` format where 1234 is the migration version
and name can be any alphanumeric and underscore combination. Migration files
are sorted based on their lowercase name. Each migration file must define a
public ``migrate`` function that accepts a ``MigrationContext`` as its argument.

When executing a migration, the migration file is imported as a module and the
``migrate`` function is executed. Renku checks project's metadata version and
applies any migration that has a higher version to the project.
"""

import importlib
import re
import shutil
from pathlib import Path
from typing import Optional, Tuple

from packaging.version import Version

from renku.command.command_builder.command import inject
from renku.core.constant import RENKU_TMP
from renku.core.errors import (
    DockerfileUpdateError,
    MigrationError,
    MigrationRequired,
    ProjectNotSupported,
    TemplateUpdateError,
)
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.migration.models.migration import MigrationContext, MigrationType
from renku.core.migration.utils import is_using_temporary_datasets_path, read_project_version
from renku.core.template.usecase import calculate_dockerfile_checksum, update_dockerfile_checksum
from renku.core.util import communication
from renku.core.util.metadata import (
    is_renku_project,
    read_renku_version_from_dockerfile,
    replace_renku_version_in_dockerfile,
)
from renku.domain_model.project import ProjectTemplateMetadata
from renku.domain_model.project_context import project_context

try:
    import importlib_resources
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore

SUPPORTED_PROJECT_VERSION = 10


def check_for_migration():
    """Checks if migration is required."""
    if is_migration_required():
        raise MigrationRequired
    elif is_project_unsupported():
        raise ProjectNotSupported


def is_migration_required():
    """Check if project requires migration."""
    return is_renku_project() and get_project_version() < SUPPORTED_PROJECT_VERSION


def is_project_unsupported():
    """Check if this version of Renku cannot work with the project."""
    return is_renku_project() and get_project_version() > SUPPORTED_PROJECT_VERSION


def is_docker_update_possible() -> bool:
    """Check if the Dockerfile can be updated to a new version of renku-python."""
    return update_dockerfile(check_only=True)[0]


@inject.autoparams("project_gateway")
def migrate_project(
    project_gateway: IProjectGateway,
    force_template_update=False,
    skip_template_update=False,
    skip_docker_update=False,
    skip_migrations=False,
    project_version=None,
    max_version=None,
    strict=False,
    migration_type: MigrationType = MigrationType.ALL,
    preserve_identifiers=False,
):
    """Migrate all project's entities.

    NOTE: The project path must be pushed to the project_context before calling this function.

    Args:
        project_gateway(IProjectGateway): The injected project gateway.
        force_template_update: Whether to force update the template  (Default value = False).
        skip_template_update: Whether to skip updating the template (Default value = False).
        skip_docker_update: Whether to skip updating the Dockerfile (Default value = False).
        skip_migrations: Whether to skip migrating project metadata (Default value = False).
        project_version: Starting migration version (Default value = False).
        max_version: Apply migration up to the given version (Default value = False).
        strict: Whether to fail on errors (Default value = False).
        migration_type(MigrationType): Type of migration to perform (Default value = MigrationType.ALL).
        preserve_identifiers: Whether to preserve ids when migrating metadata (Default value = False).

    Returns:
        Dictionary of project migration status.
    """
    template_updated = docker_updated = False
    if not is_renku_project():
        return False, template_updated, docker_updated

    try:
        project = project_context.project
    except ValueError:
        project = None

    if (
        not skip_template_update
        and project
        and hasattr(project, "template_metadata")
        and isinstance(project.template_metadata, ProjectTemplateMetadata)
        and project.template_metadata.template_source
    ):
        try:
            template_updated = _update_template()
        except TemplateUpdateError:
            raise
        except Exception as e:
            raise TemplateUpdateError("Couldn't update from template.") from e

    if not skip_docker_update:
        try:
            docker_updated, _, _ = update_dockerfile()
        except DockerfileUpdateError:
            raise
        except Exception as e:
            raise DockerfileUpdateError("Couldn't update renku version in Dockerfile.") from e

    if skip_migrations:
        return False, template_updated, docker_updated

    project_version = project_version or get_project_version()
    n_migrations_executed = 0

    migration_context = MigrationContext(strict=strict, type=migration_type, preserve_identifiers=preserve_identifiers)

    version = 1
    for version, path in get_migrations():
        if max_version and version > max_version:
            break
        if version > project_version:
            module = importlib.import_module(path)
            module_name = module.__name__.split(".")[-1]
            communication.echo(f"Applying migration {module_name}...")
            try:
                module.migrate(migration_context)
            except (Exception, BaseException) as e:
                raise MigrationError("Couldn't execute migration") from e
            n_migrations_executed += 1
    if not is_using_temporary_datasets_path():
        if n_migrations_executed > 0:
            project_context.project.version = str(version)
            project_gateway.update_project(project_context.project)

            communication.echo(f"Successfully applied {n_migrations_executed} migrations.")

        _remove_untracked_renku_files(metadata_path=project_context.metadata_path)

    return n_migrations_executed != 0, template_updated, docker_updated


def _remove_untracked_renku_files(metadata_path):
    from renku.core.constant import CACHE

    untracked_paths = [RENKU_TMP, CACHE, "vendors"]
    for path in untracked_paths:
        path = metadata_path / path
        shutil.rmtree(path, ignore_errors=True)


def _update_template() -> bool:
    """Update local files from the remote template."""
    from renku.core.template.usecase import update_template

    try:
        project = project_context.project
    except ValueError:
        # NOTE: Old project, we don't know the status until it is migrated
        return False

    if not hasattr(project, "template_metadata") or not project.template_metadata.template_version:
        return False

    return bool(update_template(interactive=False, force=False, dry_run=False))


def update_dockerfile(*, check_only=False) -> Tuple[bool, Optional[bool], Optional[str]]:
    """Update the dockerfile to the newest version of renku."""
    from renku import __version__

    if not project_context.dockerfile_path.exists():
        return False, None, None

    with open(project_context.dockerfile_path) as f:
        dockerfile_content = f.read()

    docker_version = read_renku_version_from_dockerfile()
    if not docker_version:
        if check_only:
            return False, None, None
        raise DockerfileUpdateError(
            "Couldn't update renku-python version in Dockerfile, as it doesn't contain an 'ARG RENKU_VERSION=...' line."
        )

    current_version = Version(Version(__version__).base_version)
    if Version(docker_version.base_version) >= current_version:
        return True, False, str(docker_version)

    if check_only:
        return True, True, str(docker_version)

    communication.echo("Updating dockerfile...")

    new_content = replace_renku_version_in_dockerfile(dockerfile_content=dockerfile_content, version=__version__)
    new_checksum = calculate_dockerfile_checksum(dockerfile_content=new_content)

    try:
        update_dockerfile_checksum(new_checksum=new_checksum)
    except DockerfileUpdateError:
        pass

    with open(project_context.dockerfile_path, "w") as f:
        f.write(new_content)

    communication.echo("Updated dockerfile.")

    return True, False, str(current_version)


def get_project_version():
    """Get the metadata version the renku project is on."""
    try:
        return int(read_project_version())
    except ValueError:
        return 1


def get_migrations():
    """Return a sorted list of versions and migration modules."""
    migrations = []
    for entry in importlib_resources.files("renku.core.migration").iterdir():
        match = re.search(r"^m_([0-9]{4})__[a-zA-Z0-9_-]*.py$", entry.name)

        if match is None:  # migration files match m_0000__[name].py format
            continue

        version = int(match.groups()[0])
        path = f"renku.core.migration.{Path(entry.name).stem}"
        migrations.append((version, path))

    migrations = sorted(migrations, key=lambda v: v[1].lower())
    return migrations



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset/DatasetFile view model."""

from typing import Optional

from renku.domain_model.dataset import Dataset, DatasetFile


class DatasetViewModel:
    """A view model for a ``Dataset``."""

    def __init__(self, slug: str, same_as: Optional[str]):
        self.slug: str = slug
        self.same_as: Optional[str] = same_as

    @classmethod
    def from_dataset(cls, dataset: Dataset) -> "DatasetViewModel":
        """Create view model from ``Dataset``."""
        return cls(slug=dataset.slug, same_as=dataset.same_as.value if dataset.same_as else None)


class DatasetFileViewModel:
    """A view model for a ``DatasetFile``."""

    def __init__(self, path: str, external: bool, deleted: bool, source: Optional[str], dataset: DatasetViewModel):
        self.path: str = path
        self.source: Optional[str] = source
        self.external: bool = external
        self.deleted: bool = deleted
        self.dataset: DatasetViewModel = dataset

    @classmethod
    def from_dataset_file(cls, dataset_file: DatasetFile, dataset: Dataset) -> "DatasetFileViewModel":
        """Create view model from ``DatasetFile``."""
        return cls(
            path=dataset_file.entity.path,
            source=dataset_file.source,
            external=dataset_file.is_external,
            deleted=dataset_file.date_removed is not None,
            dataset=DatasetViewModel.from_dataset(dataset),
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project initialization logic."""

import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import uuid4

from pydantic import ConfigDict, validate_call

from renku.command.command_builder.command import inject
from renku.command.mergetool import setup_mergetool
from renku.core import errors
from renku.core.config import set_value
from renku.core.constant import DATA_DIR_CONFIG_KEY, RENKU_HOME
from renku.core.git import with_worktree
from renku.core.githooks import install_githooks
from renku.core.image import ImageObjectRequest
from renku.core.interface.database_gateway import IDatabaseGateway
from renku.core.migration.utils import OLD_METADATA_PATH
from renku.core.project import set_project_image
from renku.core.storage import init_external_storage, storage_installed
from renku.core.template.template import (
    FileAction,
    RenderedTemplate,
    TemplateAction,
    copy_template_to_project,
    fetch_templates_source,
    get_file_actions,
    set_template_parameters,
)
from renku.core.template.usecase import select_template
from renku.core.util import communication
from renku.core.util.contexts import with_project_metadata
from renku.core.util.git import with_commit
from renku.core.util.os import is_path_empty
from renku.domain_model.project import Project
from renku.domain_model.project_context import project_context
from renku.domain_model.template import Template, TemplateMetadata
from renku.version import __version__, is_release

if TYPE_CHECKING:
    from renku.infrastructure.repository import Repository


def create_backup_branch(path):
    """Creates a backup branch of the repository.

    Args:
        path: Repository path.

    Returns:
        Name of the backup branch.
    """
    repository = project_context.repository

    branch_name = None
    if not is_path_empty(path):
        try:
            if repository.head.is_valid():
                hexsha = repository.head.commit.hexsha[:7]

                branch_name = f"pre_renku_init_{hexsha}"

                for ref in repository.branches:
                    if branch_name == ref.name:
                        branch_name = f"pre_renku_init_{hexsha}_{uuid4().hex}"
                        break

                with with_worktree(
                    branch_name=branch_name,
                    commit=repository.head.commit,
                    merge_args=["--no-ff", "-s", "recursive", "-X", "ours", "--allow-unrelated-histories"],
                ):
                    communication.warn(f"Saving current data in branch {branch_name}")
        except AttributeError:
            communication.echo("Warning! Overwriting non-empty folder.")
        except errors.GitCommandError:
            raise

    return branch_name


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def init_project(
    external_storage_requested: bool,
    path: str,
    name: Optional[str],
    description: Optional[str],
    keywords: Optional[List[str]],
    image_request: Optional[ImageObjectRequest],
    template_id: Optional[str],
    template_source: Optional[str],
    template_ref: Optional[str],
    input_parameters: Dict[str, str],
    custom_metadata: Optional[Dict[str, Any]],
    force: bool,
    data_dir: Optional[Path],
    initial_branch: Optional[str],
    install_mergetool: bool,
):
    """Initialize a renku project.

    Args:
        external_storage_requested: Whether external storage should be used.
        path: Path to initialize repository at.
        name: Name of the project.
        description: Description of the project.
        keywords: keywords for the project.
        image_request(Optional[ImageObjectRequest]): Project's image.
        template_id: id of the template to use.
        template_source: Source to get the template from.
        template_ref: Reference to use to get the template.
        input_parameters: Template parameters.
        custom_metadata: Custom JSON-LD metadata for project.
        force: Whether to overwrite existing files and delete existing metadata.
        data_dir: Where to store dataset data.
        initial_branch: Default git branch.
        install_mergetool(bool): Whether to set up the renku metadata mergetool in the created project.
    """
    if not project_context.external_storage_requested:
        external_storage_requested = False

    project_context.push_path(path, save_changes=True)
    if data_dir is not None:
        project_context.datadir = str(data_dir)
    project_context.external_storage_requested = external_storage_requested

    communication.echo("Initializing Git repository...")
    project_context.repository = init_repository(force=force, user=None, initial_branch=initial_branch)

    # Initialize an empty database
    database_gateway = inject.instance(IDatabaseGateway)
    database_gateway.initialize()

    templates_source = fetch_templates_source(source=template_source, reference=template_ref)
    template = select_template(templates_source=templates_source, id=template_id)

    if template is None:
        raise errors.TemplateNotFoundError(f"Couldn't find template with id {template_id}")

    namespace, name = Project.get_namespace_and_name(
        remote=project_context.remote, name=name, repository=project_context.repository
    )
    name = name or os.path.basename(path.rstrip(os.path.sep))

    metadata: Dict[str, Any] = dict()
    # NOTE: supply metadata
    metadata["__template_source__"] = template_source
    metadata["__template_ref__"] = template_ref
    metadata["__template_id__"] = template_id
    metadata["__namespace__"] = namespace or ""
    metadata["__sanitized_project_name__"] = ""
    metadata["__repository__"] = ""
    metadata["__project_slug__"] = ""
    metadata["__project_description__"] = description
    if is_release() and "__renku_version__" not in metadata:
        metadata["__renku_version__"] = __version__
    metadata["__name__"] = name
    metadata["name"] = name  # NOTE: kept for backwards compatibility
    metadata["__template_version__"] = template.version
    metadata["__automated_update__"] = True  # TODO: This should come from a command line flag

    template_metadata = TemplateMetadata.from_dict(metadata=metadata)
    template_metadata.update(template=template)
    # TODO: Validate input_parameters to make sure they don't contain __\w+__ keys
    set_template_parameters(template=template, template_metadata=template_metadata, input_parameters=input_parameters)

    rendered_template = template.render(metadata=template_metadata)
    actions = get_file_actions(
        rendered_template=rendered_template, template_action=TemplateAction.INITIALIZE, interactive=False
    )

    if not force:
        appends = [k for k, v in actions.items() if v == FileAction.APPEND]
        overwrites = [k for k, v in actions.items() if v == FileAction.OVERWRITE]

        message = ""
        if overwrites:
            overwrites_str = "\n\t".join(sorted(overwrites))
            message += f"The following files exist in the directory and will be overwritten:\n\t{overwrites_str}\n"
        if appends:
            appends_str = "\n\t".join(sorted(appends))
            message += f"The following files exist in the directory and will be appended to:\n\t{appends_str}\n"
        if message:
            communication.confirm(f"{message}Proceed?", abort=True, warning=True)

    branch_name = create_backup_branch(path=path)

    # NOTE: clone the repo
    communication.echo("Initializing new Renku repository... ")
    with project_context.lock:
        try:
            create_from_template(
                rendered_template=rendered_template,
                actions=actions,
                name=name,
                custom_metadata=custom_metadata,
                data_dir=data_dir,
                description=description,
                keywords=keywords,
                install_mergetool=install_mergetool,
                image_request=image_request,
            )
        except FileExistsError as e:
            raise errors.InvalidFileOperation(e)

    # NOTE: Installing githooks
    warning_messages = install_githooks(force=force, path=project_context.path)
    if warning_messages:
        for message in warning_messages:
            communication.warn(message)

    if branch_name:
        communication.echo(
            f"Project initialized.\nYou can undo this command by running 'git reset --hard {branch_name}'\n"
            f"You can see changes made by running 'git diff {branch_name} HEAD'"
        )
    else:
        communication.echo("Project initialized.")


def init_repository(force=False, user=None, initial_branch=None) -> "Repository":
    """Initialize an empty Renku repository."""
    from renku.infrastructure.repository import Repository

    # initialize repo and set user data
    path = project_context.path
    if force and (path / RENKU_HOME).exists():
        shutil.rmtree(path / RENKU_HOME)
    repository = Repository.initialize(path=path, branch=initial_branch)
    if user:
        with repository.get_configuration(writable=True) as config_writer:
            for key, value in user.items():
                config_writer.set_value("user", key, value)

    # verify if git user information is available
    _ = repository.get_user()

    # initialize LFS if it is requested and installed
    if project_context.external_storage_requested and storage_installed():
        init_external_storage(force=force)

    return repository


def create_from_template(
    rendered_template: RenderedTemplate,
    actions: Dict[str, FileAction],
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    custom_metadata: Optional[Dict] = None,
    data_dir: Optional[Union[str, Path]] = None,
    commit_message: Optional[str] = None,
    description: Optional[str] = None,
    keywords: Optional[List[str]] = None,
    image_request: Optional[ImageObjectRequest] = None,
    install_mergetool: bool = False,
):
    """Initialize a new project from a template.

    Args:
        rendered_template(RenderedTemplate): Rendered template.
        actions(Dict[str, FileAction]): mapping of paths and actions to take.
        name(Optional[str]): Name of the project (Default value = None).
        namespace(Optional[str]): Namespace of the project (Default value = None).
        custom_metadata(Optional[Dict]): Custom JSON-LD metadata (Default value = None).
        data_dir(Optional[str]): Where to store dataset data (Default value = None).
        commit_message(Optional[str]): Message for initial commit (Default value = None).
        description(Optional[str]): Description of the project (Default value = None).
        keywords(Optional[List[str]]): Keywords for project (Default value = None).
        image_request(Optional[ImageObjectRequest]): Project's image (Default value = None).
        install_mergetool(bool): Whether to set up renku metadata mergetool (Default value = False).
    """
    commit_only = [f"{RENKU_HOME}/", str(project_context.template_checksums_path)] + list(rendered_template.get_files())

    if install_mergetool:
        commit_only.append(".gitattributes")

    if data_dir:
        data_path = project_context.path / data_dir
        data_path.mkdir(parents=True, exist_ok=True)
        keep = data_path / ".gitkeep"
        keep.touch(exist_ok=True)
        commit_only.append(str(keep))

    # add metadata.yml for backwards compatibility
    metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
    with open(metadata_path, "w") as f:
        f.write(
            "# Dummy file kept for backwards compatibility, does not contain actual version\n"
            "'http://schema.org/schemaVersion': '9'"
        )

    with with_commit(
        repository=project_context.repository,
        transaction_id=project_context.transaction_id,
        commit_message=commit_message,
        commit_only=commit_only,
        skip_dirty_checks=True,
    ):
        with with_project_metadata(
            name=name, namespace=namespace, description=description, custom_metadata=custom_metadata, keywords=keywords
        ) as project:
            copy_template_to_project(rendered_template=rendered_template, project=project, actions=actions)

            # NOTE: Copy image to project
            set_project_image(image_request=image_request)

        if install_mergetool:
            setup_mergetool()

        if data_dir:
            set_value("renku", DATA_DIR_CONFIG_KEY, str(data_dir))


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def create_from_template_local(
    template_path: Path,
    name: str,
    namespace: str,
    metadata: Optional[Dict] = None,
    custom_metadata: Optional[Dict] = None,
    default_metadata: Optional[Dict] = None,
    template_version: Optional[str] = None,
    immutable_template_files: Optional[List[str]] = None,
    automated_template_update: bool = True,
    user: Optional[Dict[str, str]] = None,
    initial_branch: Optional[str] = None,
    commit_message: Optional[str] = None,
    description: Optional[str] = None,
    keywords: Optional[List[str]] = None,
    data_dir: Optional[str] = None,
    ssh_supported: bool = False,
    image_request: Optional[ImageObjectRequest] = None,
):
    """Initialize a new project from a template.

    Args:
        template_path(Path): Path to template.
        name(str): project name.
        namespace(str): project namespace.
        metadata(Optional[Dict]): Project metadata (Default value = None).
        custom_metadata(Optional[Dict]): Custom JSON-LD metadata (Default value = None).
        default_metadata(Optional[Dict]): Default project metadata (Default value = None).
        template_version(Optional[str]): Version of the template (Default value = None).
        immutable_template_files(Optional[List[str]]): Immutable template files (Default value = None).
        automated_template_update(bool): If template can be updated automatically (Default value = True).
        user(Optional[Dict[str, str]]): Git user (Default value = None).
        initial_branch(Optional[str]): Name of initial/main branch (Default value = None).
        commit_message(Optional[str]): Message of initial commit (Default value = None).
        description(Optional[str]): Project description (Default value = None).
        keywords(Optional[List[str]]): Project keywords (Default value = None).
        data_dir(Optional[str]): Project base data directory (Default value = None).
        ssh_supported(Optional[bool]): Whether the project supports ssh (Default value = False).
        image_request(Optional[ImageObjectRequest]): Project's image (Default value = None).
    """
    metadata = metadata or {}
    default_metadata = default_metadata or {}

    metadata = {**default_metadata, **metadata}

    project_context.repository = init_repository(force=False, user=user, initial_branch=initial_branch)

    # Initialize an empty database
    database_gateway = inject.instance(IDatabaseGateway)
    database_gateway.initialize()

    if "__name__" not in metadata:
        metadata["name"] = name
        metadata["__name__"] = name

    metadata["__template_version__"] = template_version

    template = Template(
        id=metadata["__template_id__"],
        aliases=[],
        name="",
        description="",
        parameters={},
        icon="",
        ssh_supported=ssh_supported,
        immutable_files=immutable_template_files or [],
        allow_update=automated_template_update,
        source=metadata["__template_source__"],
        reference=metadata["__template_ref__"],
        version=template_version,
        path=template_path,
        templates_source=None,
    )

    template_metadata = TemplateMetadata.from_dict(metadata=metadata)
    template_metadata.update(template=template)
    set_template_parameters(template=template, template_metadata=template_metadata, input_parameters={})

    rendered_template = template.render(metadata=template_metadata)
    actions = get_file_actions(
        rendered_template=rendered_template, template_action=TemplateAction.INITIALIZE, interactive=False
    )

    create_from_template(
        rendered_template=rendered_template,
        actions=actions,
        name=name,
        namespace=namespace,
        custom_metadata=custom_metadata,
        commit_message=commit_message,
        description=description,
        keywords=keywords,
        data_dir=data_dir,
        image_request=image_request,
    )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache serializers."""

from marshmallow import Schema, fields

from renku.ui.service.serializers.cache import DockerfileStatusResponse, ProjectCompatibilityResponse
from renku.ui.service.serializers.common import RenkuSyncSchema
from renku.ui.service.serializers.rpc import JsonRPCResponse


class ProjectMigrateResponse_1_0(RenkuSyncSchema):
    """Response schema for project migrate."""

    was_migrated = fields.Boolean()
    template_migrated = fields.Boolean()
    docker_migrated = fields.Boolean()
    messages = fields.List(fields.String)


class ProjectMigrateResponseRPC_1_0(JsonRPCResponse):
    """RPC response schema for project migrate."""

    result = fields.Nested(ProjectMigrateResponse_1_0)


class TemplateStatusResponse_1_5(Schema):
    """Response schema outlining template status for migrations check."""

    automated_template_update = fields.Boolean(
        metadata={"description": "Whether or not the project template explicitly supports automated updates."}
    )
    newer_template_available = fields.Boolean(
        metadata={
            "description": "Whether or not the current version of the project template differs from the "
            "one used in the project."
        }
    )
    template_source = fields.String(
        metadata={
            "description": "Source of the template repository, "
            "either a Git URL or 'renku' if an embedded template was used."
        }
    )
    template_ref = fields.String(
        metadata={
            "description": "The branch/tag/commit from the template_source repository "
            "that was used to create this project."
        }
    )
    template_id = fields.String(metadata={"description": "The id of the template in the template repository."})

    project_template_version = fields.String(
        allow_none=True, metadata={"description": "The version of the template last used in the user's project."}
    )
    latest_template_version = fields.String(
        allow_none=True, metadata={"description": "The current version of the template in the template repository."}
    )


class ProjectMigrationCheckResponse_1_5(Schema):
    """Response schema for project migration check."""

    project_supported = fields.Boolean(
        metadata={
            "description": "Determines whether this project is a Renku project that is supported by the version "
            "running on this service (not made with a newer version)."
        }
    )
    core_renku_version = fields.String(metadata={"description": "Version of Renku running in this service."})
    project_renku_version = fields.String(metadata={"description": "Version of Renku last used to change the project."})

    core_compatibility_status = fields.Nested(
        ProjectCompatibilityResponse,
        metadata={"description": "Fields detailing the compatibility of the project with this core service version."},
    )
    dockerfile_renku_status = fields.Nested(
        DockerfileStatusResponse,
        metadata={"description": "Fields detailing the status of the Dockerfile in the project."},
    )
    template_status = fields.Nested(
        TemplateStatusResponse_1_5,
        metadata={"description": "Fields detailing the status of the project template used by this project."},
    )


class ProjectMigrationCheckResponseRPC_1_5(JsonRPCResponse):
    """RPC response schema for project migration check."""

    result = fields.Nested(ProjectMigrationCheckResponse_1_5)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate project to the latest Renku version."""

from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

from pydantic import ConfigDict, validate_call

from renku.command.command_builder.command import Command
from renku.core.errors import MinimumVersionError
from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION
from renku.domain_model.project_context import project_context

SUPPORTED_RENKU_PROJECT = 1
MIGRATION_REQUIRED = 2
UNSUPPORTED_PROJECT = 4
NON_RENKU_REPOSITORY = 8
TEMPLATE_UPDATE_POSSIBLE = 16
AUTOMATED_TEMPLATE_UPDATE_SUPPORTED = 32
DOCKERFILE_UPDATE_POSSIBLE = 64


@dataclass
class CoreStatusResult:
    """Core migration status."""

    migration_required: bool
    project_metadata_version: Optional[int]
    current_metadata_version: int


@dataclass
class DockerfileStatusResult:
    """Docker migration status."""

    automated_dockerfile_update: bool
    newer_renku_available: Optional[bool]
    dockerfile_renku_version: Optional[str]
    latest_renku_version: str


@dataclass
class TemplateStatusResult:
    """Template migration status."""

    automated_template_update: bool
    newer_template_available: bool
    project_template_version: Optional[str]
    latest_template_version: Optional[str]
    template_source: Optional[str]
    template_ref: Optional[str]
    template_id: Optional[str]
    ssh_supported: bool


@dataclass
class MigrationCheckResult:
    """Migration check output."""

    project_supported: bool
    core_renku_version: str
    project_renku_version: Optional[str]
    core_compatibility_status: Union[CoreStatusResult, Exception]
    dockerfile_renku_status: Union[DockerfileStatusResult, Exception]
    template_status: Union[TemplateStatusResult, Exception]

    @staticmethod
    def from_minimum_version_error(minimum_version_error: MinimumVersionError) -> "MigrationCheckResult":
        """Create a migration check when the project isn't supported yet."""
        from renku import __version__

        return MigrationCheckResult(
            project_supported=False,
            core_renku_version=str(minimum_version_error.current_version),
            project_renku_version=f">={minimum_version_error.minimum_version}",
            core_compatibility_status=CoreStatusResult(
                migration_required=False,
                project_metadata_version=None,
                current_metadata_version=SUPPORTED_PROJECT_VERSION,
            ),
            dockerfile_renku_status=DockerfileStatusResult(
                dockerfile_renku_version="unknown",
                latest_renku_version=__version__,
                newer_renku_available=False,
                automated_dockerfile_update=False,
            ),
            template_status=TemplateStatusResult(
                automated_template_update=False,
                newer_template_available=False,
                template_source="unknown",
                template_ref="unknown",
                template_id="unknown",
                project_template_version="unknown",
                latest_template_version="unknown",
                ssh_supported=False,
            ),
        )


def migrations_check():
    """Return a command for a migrations check."""
    return Command().command(_migrations_check).with_database(write=False)


def _migrations_check() -> MigrationCheckResult:
    """Check migration status of project.

    Returns:
        Dictionary of project migrations, template and dockerfile status.
    """
    from renku.core.migration.migrate import is_project_unsupported

    core_version, latest_version = _migrations_versions()

    try:
        core_compatibility_status: Union[CoreStatusResult, Exception] = _metadata_migration_check()
    except Exception as e:
        core_compatibility_status = e

    try:
        docker_status: Union[DockerfileStatusResult, Exception] = _dockerfile_migration_check()
    except Exception as e:
        docker_status = e

    try:
        template_status: Union[TemplateStatusResult, Exception] = _template_migration_check()
    except Exception as e:
        template_status = e

    return MigrationCheckResult(
        project_supported=not is_project_unsupported(),
        core_renku_version=core_version,
        project_renku_version=latest_version,
        core_compatibility_status=core_compatibility_status,
        dockerfile_renku_status=docker_status,
        template_status=template_status,
    )


def migrations_versions():
    """Return a command to get source and destination migration versions."""
    return Command().command(_migrations_versions).lock_project().with_database()


def _migrations_versions() -> Tuple[str, Optional[str]]:
    """Return source and destination migration versions.

    Returns:
        Tuple of current version and project version.
    """
    from renku import __version__

    try:
        latest_agent = project_context.latest_agent
    except ValueError:
        # NOTE: maybe old project
        from renku.core.migration.utils import read_latest_agent

        latest_agent = read_latest_agent()

    return __version__, latest_agent


def _template_migration_check() -> TemplateStatusResult:
    """Return template migration status.

    Returns:
        Dictionary of template migration status.
    """
    from renku.core.config import get_value
    from renku.core.template.usecase import check_for_template_update

    project = project_context.project
    template_source = project.template_metadata.template_source
    template_ref = project.template_metadata.template_ref
    template_id = project.template_metadata.template_id
    ssh_supported = project.template_metadata.ssh_supported

    ssh_supported = get_value("renku", "ssh_supported") == "true" or ssh_supported

    update_available, update_allowed, current_version, new_version = check_for_template_update(project)

    return TemplateStatusResult(
        automated_template_update=update_allowed,
        newer_template_available=update_available,
        project_template_version=current_version,
        latest_template_version=new_version,
        template_source=template_source,
        template_ref=template_ref,
        template_id=template_id,
        ssh_supported=ssh_supported,
    )


def dockerfile_migration_check():
    """Return a command for a Dockerfile migrations check."""
    return Command().command(_dockerfile_migration_check)


def _dockerfile_migration_check() -> DockerfileStatusResult:
    """Return Dockerfile migration status.

    Returns:
        Dictionary of Dockerfile migration status.
    """
    from renku import __version__
    from renku.core.migration.migrate import update_dockerfile

    automated_dockerfile_update, newer_renku_available, dockerfile_renku_version = update_dockerfile(check_only=True)

    return DockerfileStatusResult(
        automated_dockerfile_update=automated_dockerfile_update,
        newer_renku_available=newer_renku_available,
        dockerfile_renku_version=dockerfile_renku_version,
        latest_renku_version=__version__,
    )


def metadata_migration_check():
    """Return a command for a metadata migrations check."""
    return Command().command(_metadata_migration_check)


def _metadata_migration_check() -> CoreStatusResult:
    """Return metadata migration status.

    Returns:
        Dictionary of metadata migration status.
    """
    from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION, get_project_version, is_migration_required

    return CoreStatusResult(
        migration_required=is_migration_required(),
        project_metadata_version=get_project_version(),
        current_metadata_version=SUPPORTED_PROJECT_VERSION,
    )


def migrate_project_command():
    """Return a command to migrate all project's entities."""
    from renku.core.migration.migrate import migrate_project

    return Command().command(migrate_project).lock_project().require_clean().with_database(write=True)


def check_project():
    """Return a command to check if repository is a renku project, unsupported, or requires migration."""
    return Command().command(_check_project).with_database(write=False)


def _check_project():
    from renku.core.migration.migrate import is_docker_update_possible, is_migration_required, is_project_unsupported
    from renku.core.template.usecase import check_for_template_update
    from renku.core.util.metadata import is_renku_project

    if not is_renku_project():
        return NON_RENKU_REPOSITORY
    elif is_project_unsupported():
        return UNSUPPORTED_PROJECT

    try:
        _ = project_context.project
    except ValueError:
        return MIGRATION_REQUIRED
    else:
        if hasattr(project_context.project, "template_source"):
            # NOTE: v10 migration not done
            return MIGRATION_REQUIRED

    # NOTE: ``project.automated_update`` is deprecated. We always allow template update for a project
    status = AUTOMATED_TEMPLATE_UPDATE_SUPPORTED

    if check_for_template_update(project_context.project)[0]:
        status |= TEMPLATE_UPDATE_POSSIBLE
    if is_docker_update_possible():
        status |= DOCKERFILE_UPDATE_POSSIBLE
    if is_migration_required():
        return status | MIGRATION_REQUIRED

    return status | SUPPORTED_RENKU_PROJECT


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _check_immutable_template_files(paths: List[str]):
    """Check paths and return a list of those that are marked immutable in the project template.

    Args:
        paths(List[str]): Paths to check.

    Returns:
        List of immutable template files.
    """
    immutable_template_files = project_context.project.template_metadata.immutable_template_files or []

    return [p for p in paths if str(p) in immutable_template_files]


def check_immutable_template_files_command():
    """Command for checking immutable template files."""
    return Command().command(_check_immutable_template_files).with_database()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Various metadata migrations for v10."""

import io
import json
import os
from datetime import timezone
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Tuple, cast

import zstandard as zstd

from renku.command.checks.activities import fix_activity_dates
from renku.command.checks.workflow import fix_plan_dates
from renku.command.command_builder import inject
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.migration.models import v10 as old_schema
from renku.core.migration.models.migration import MigrationContext, MigrationType
from renku.core.util import communication
from renku.core.util.datetime8601 import local_now
from renku.domain_model.dataset import Dataset
from renku.domain_model.project import Project, ProjectTemplateMetadata
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity
from renku.domain_model.workflow.plan import AbstractPlan


def migrate(migration_context: MigrationContext):
    """Migration function."""
    from renku.infrastructure.gateway.activity_gateway import reindex_catalog

    migrate_project_template_data()
    migrate_old_metadata_namespaces()

    if MigrationType.WORKFLOWS in migration_context.options.type:
        migrate_activity_ids()
        fix_plan_times()
        fix_activity_times()

    migrate_remote_entity_ids()
    fix_dataset_date_modified()
    fix_dataset_image_ids()
    fix_removed_plans()

    # NOTE: Rebuild all workflow catalogs since ids and times have changed
    communication.echo("Rebuilding workflow metadata")
    reindex_catalog(project_context.database)


def migrate_old_metadata_namespaces():
    """Migrate old namespaces (e.g. renku.cli.* -> renku.ui.cli.*)."""
    communication.echo("Migrating old namespaces")
    compressor = zstd.ZstdCompressor()
    decompressor = zstd.ZstdDecompressor()
    for root, _, files in os.walk(project_context.database_path):
        root = Path(root)
        for file_name in files:
            path = root / file_name
            compressed = False
            with open(path, "rb") as file:
                header = int.from_bytes(file.read(4), "little")
                file.seek(0)
                if header == zstd.MAGIC_NUMBER:
                    with decompressor.stream_reader(file) as compressed_file:
                        data = json.load(compressed_file)
                        compressed = True
                else:
                    data = json.load(file)

            # modify namespace
            transformation = [
                ("renku.core.models.", "renku.domain_model."),
                ("renku.core.metadata.", "renku.infrastructure."),
            ]
            nested_update(data, target_key="@renku_data_type", transforms=transformation)

            if compressed:
                with open(path, "wb") as fb, compressor.stream_writer(fb) as compression_writer:
                    with io.TextIOWrapper(compression_writer) as out:
                        json.dump(data, out, ensure_ascii=False)
            else:
                with open(path, "w") as ft:
                    json.dump(data, ft, ensure_ascii=False, sort_keys=True, indent=2)


def nested_update(data: Dict[str, Any], target_key: str, transforms: List[Tuple[str, str]]) -> None:
    """Update a key's value based on transformations (from, to) in a deeply nested dictionary."""
    for k in list(data.keys()):
        value = data[k]
        if isinstance(value, str) and k == target_key:
            for lookup, replacement in transforms:
                value = value.replace(lookup, replacement)
            data[k] = value
        elif isinstance(value, dict):
            nested_update(value, target_key=target_key, transforms=transforms)
        elif isinstance(value, list) and len(value) > 0 and any(isinstance(v, dict) for v in value):
            result = []
            for v in value:
                if isinstance(v, dict):
                    nested_update(v, target_key=target_key, transforms=transforms)

                result.append(v)
            data[k] = result


def migrate_activity_ids():
    """Fix activity related ids.

    Fixes ids not starting with `/activities/`.
    Fixes generations having `/usages/` in their id.
    """
    communication.echo("Fixing activity ids")

    database = project_context.database

    activities: List[Activity] = list(database["activities"].values())

    def fix_id(id: str, changed: bool) -> Tuple[str, bool]:
        """Fix an activity related id."""
        if not id.startswith("/activities/"):
            return f"/activities/{id.lstrip('/')}", True
        return id, changed

    for activity in activities:
        changed = False
        activity.unfreeze()
        old_activity_id = activity.id
        activity.id, changed = fix_id(activity.id, changed)
        activity.association.id, changed = fix_id(activity.association.id, changed)

        for generation in activity.generations:
            id, changed = fix_id(generation.id, changed)

            if "/usages/" in generation.id:
                id = generation.id.replace("/usages/", "/generations/")
                changed = True
            object.__setattr__(generation, "id", id)

        for entity in chain(activity.usages, activity.parameters):
            id, changed = fix_id(entity.id, changed)
            object.__setattr__(entity, "id", id)

        if changed:
            database["activities"].pop(old_activity_id)
            activity._p_changed = True
            old_id = str(activity._p_oid)
            os.unlink(project_context.database_path / old_id[0:2] / old_id[2:4] / old_id)
            activity.reassign_oid()
            database["activities"].add(activity)
        activity.freeze()
    database["activities"]._p_changed = True
    database.commit()


def migrate_remote_entity_ids():
    """Change `remote-entity` to `remote-entities` in ids."""
    database = project_context.database

    def fix_dataset_files_based_on(dataset):
        changed = False
        for file in dataset.dataset_files:
            if file.based_on is not None:
                file.based_on.id = file.based_on.id.replace("/remote-entity//", "/remote-entities/")
                file.based_on.id = file.based_on.id.replace("/remote-entity/", "/remote-entities/")
                changed = True

        if changed:
            dataset._p_changed = True

    datasets: List[Dataset] = list(database["datasets-provenance-tails"].values())

    for dataset in datasets:
        fix_dataset_files_based_on(dataset)

        while dataset.derived_from is not None:
            dataset = database.get_by_id(id=dataset.derived_from.url_id)
            fix_dataset_files_based_on(dataset)

    database.commit()


@inject.autoparams()
def migrate_project_template_data(project_gateway: IProjectGateway):
    """Migrate project template metadata to nested model."""
    if not hasattr(project_context.project, "template_source"):
        # NOTE: already on newest version
        return

    old_project = cast(
        old_schema.Project,
        project_context.database.get_from_path(
            "project", override_type=f"{old_schema.Project.__module__}.{old_schema.Project.__qualname__}"
        ),
    )

    project = Project(
        agent_version=old_project.agent_version,
        annotations=old_project.annotations,
        creator=old_project.creator,
        date_created=old_project.date_created,
        description=old_project.description,
        id=old_project.id,
        name=old_project.name,
        version=old_project.version,
        keywords=old_project.keywords,
        template_metadata=ProjectTemplateMetadata(
            template_id=old_project.template_id,
            metadata=old_project.template_metadata,
            template_ref=old_project.template_ref,
            template_source=old_project.template_source,
            template_version=old_project.template_version,
            immutable_template_files=old_project.immutable_template_files,
        ),
    )

    project_gateway.update_project(project)
    project_context.database.commit()


@inject.autoparams("plan_gateway")
def fix_plan_times(plan_gateway: IPlanGateway):
    """Rename plan's date-related attributes and add timezone to invalidation time."""
    plans: List[AbstractPlan] = plan_gateway.get_all_plans()

    for plan in plans:
        plan.unfreeze()
        if hasattr(plan, "invalidated_at"):
            plan.date_removed = plan.invalidated_at
            del plan.invalidated_at
        elif not hasattr(plan, "date_removed"):
            plan.date_removed = None
        if not hasattr(plan, "date_created"):
            plan.date_created = getattr(plan, "date_modified", None) or plan.date_removed or local_now()

        if plan.date_removed is not None:
            if plan.date_removed.tzinfo is None:
                # NOTE: There was a bug that caused date_removed to be set without timezone (as UTC time)
                # so we patch in the timezone here
                plan.date_removed = plan.date_removed.astimezone(timezone.utc)
            if plan.date_removed < plan.date_created:
                # NOTE: Fix invalidation times set before creation date on plans
                plan.date_removed = plan.date_created
        plan.freeze()

    fix_plan_dates(plans=plans, plan_gateway=plan_gateway)
    project_context.database.commit()


@inject.autoparams("activity_gateway")
def fix_activity_times(activity_gateway: IActivityGateway):
    """Make sure activities have valid start/end/delete dates."""
    fix_activity_dates(activities=activity_gateway.get_all_activities(include_deleted=True))
    project_context.database.commit()


@inject.autoparams("dataset_gateway")
def fix_dataset_date_modified(dataset_gateway: IDatasetGateway):
    """Change date_created and date_modified to have correct semantics."""

    def fix_creation_date(dataset):
        """Check creation date to make sure that it's after project's creation date."""
        if dataset.date_created and dataset.date_created < project_context.project.date_created:
            dataset.unfreeze()
            try:
                dataset.date_created = min([f.date_added for f in dataset.files])
            except (ValueError, TypeError):
                dataset.date_created = project_context.project.date_created
            else:
                if dataset.date_created < project_context.project.date_created:
                    dataset.date_created = project_context.project.date_created
            dataset.freeze()

    tails = dataset_gateway.get_provenance_tails()

    for dataset_tail in tails:
        dataset = dataset_tail
        found_datasets: List[Dataset] = []
        previous_modification_date = local_now()

        while dataset.derived_from is not None:
            fix_creation_date(dataset)
            modification_date = dataset.date_removed or dataset.date_created

            if modification_date is not None:
                # NOTE: This happened in a project due to a timezone change
                if modification_date > previous_modification_date:
                    modification_date = previous_modification_date
                dataset.unfreeze()
                dataset.date_modified = modification_date
                dataset.freeze()
                previous_modification_date = modification_date
            found_datasets.append(dataset)
            dataset = dataset_gateway.get_by_id(dataset.derived_from.value)

        fix_creation_date(dataset)
        # NOTE: first dataset in chain
        modification_date = dataset.date_published or dataset.date_created
        if modification_date is not None:
            dataset.unfreeze()
            dataset.date_modified = modification_date
            dataset.freeze()

        for child in found_datasets:
            child.unfreeze()
            child.date_created = dataset.date_created
            child.date_published = dataset.date_published
            child.freeze()

    project_context.database.commit()


@inject.autoparams("dataset_gateway")
def fix_dataset_image_ids(dataset_gateway: IDatasetGateway):
    """Remove dashes from dataset image IDs."""
    for dataset in dataset_gateway.get_provenance_tails():
        while True:
            if dataset.images:
                for image in dataset.images:
                    image.id = image.id.replace("-", "")

                dataset._p_changed = True

            if not dataset.derived_from:
                break

            dataset = dataset_gateway.get_by_id(dataset.derived_from.value)

    project_context.database.commit()


@inject.autoparams("plan_gateway")
def fix_removed_plans(plan_gateway: IPlanGateway):
    """Create a derivative if a removed plan doesn't have one."""
    plans: List[AbstractPlan] = plan_gateway.get_all_plans()

    for plan in plans:
        if plan.date_removed and plan.derived_from is None:
            derived_plan = plan.derive()
            derived_plan.date_modified = plan.date_modified
            derived_plan.delete(when=plan.date_removed)
            plan_gateway.add(derived_plan)

            plan.unfreeze()
            plan.date_removed = None
            plan.freeze()

    project_context.database.commit()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Project.

Project class acts as a context for other Renku entities like Dataset, or
Inputs/Outputs. It provides access to internals of a Renku project for such
entities.

Normally, you do not need to create an instance of Project class directly
unless you want to have access to Project metadata (e.g. path) or get its status. To separate
parts of your script that uses Renku entities, you can create a Project context
manager and interact with Renku inside it:

.. code-block:: python

    from renku.api import Project, Input

    with Project():
        input_1 = Input("input_1", "path_1")

You can use Project's ``status`` method to get info about outdated outputs and
activities, and modified or deleted inputs:

.. code-block:: python

    from renku.api import Project

    outdated_generations, outdated_activities, modified_inputs, deleted_inputs = Project().status()

"""

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

from werkzeug.local import LocalStack

from renku.command.status import get_status_command
from renku.core.util.git import get_git_repository
from renku.core.workflow.run import StatusResult
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.infrastructure.repository import Repository


class Project:
    """API Project context class."""

    _project_contexts: LocalStack = LocalStack()

    def __init__(self):
        try:
            repository = get_git_repository()
            path = repository.path
        except ValueError:
            repository = None
            path = Path(".")

        project_context.replace_path(path)
        project_context.repository = repository  # type: ignore
        self._path = project_context.path
        self._repository = repository

    def __enter__(self):
        self._project_contexts.push(self)

        return self

    def __exit__(self, type, value, traceback):
        context = self._project_contexts.pop()
        if context is not self:
            raise RuntimeError("Project context was changed.")

    @property
    def repository(self) -> Optional["Repository"]:
        """Return the Repository instance."""
        return self._repository

    @property
    def path(self) -> Path:
        """Absolute path to project's root directory."""
        return self._path

    def status(self, paths: Optional[List[Union[Path, str]]] = None, ignore_deleted: bool = False) -> StatusResult:
        """Return status of a project.

        Args:
            paths(Optional[List[Union[Path, str]]]): Limit the status to this list of paths (Default value = None).
            ignore_deleted(bool): Whether to ignore deleted generations.

        Returns:
            StatusResult: Status of the project.

        """
        return get_status_command().build().execute(paths=paths, ignore_deleted=ignore_deleted).output



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache serializers for jobs."""
from marshmallow import Schema, fields

from renku.ui.service.serializers.common import CreationSchema
from renku.ui.service.serializers.rpc import JsonRPCResponse


class ProjectJobDetails(Schema):
    """Project details part of JobDetails response."""

    project_id = fields.String()

    name = fields.String()
    fullname = fields.String()
    email = fields.String()
    owner = fields.String()
    ref = fields.String()


class JobDetails(CreationSchema):
    """Response job details."""

    updated_at = fields.DateTime()

    job_id = fields.String(required=True)
    project = fields.Nested(ProjectJobDetails)
    renku_op = fields.String()

    state = fields.String()
    extras = fields.Dict()
    client_extras = fields.String()

    response = fields.Dict()


class JobListResponse(Schema):
    """Response schema for job listing."""

    jobs = fields.List(fields.Nested(JobDetails), required=True)


class JobListResponseRPC(JsonRPCResponse):
    """RPC response schema for jobs listing."""

    result = fields.Nested(JobListResponse)


class JobDetailsResponseRPC(Schema):
    """Response schema for job details."""

    result = fields.Nested(JobDetails)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check for required Git hooks."""

from io import StringIO

from renku.command.util import WARNING
from renku.core.githooks import HOOKS
from renku.core.util.git import get_hook_path
from renku.domain_model.project_context import project_context

try:
    import importlib_resources  # type: ignore[import]
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore


def check_git_hooks_installed(**_):
    """Checks if all necessary hooks are installed.

    Args:
        _: keyword arguments.

    Returns:
        Tuple of whether git hooks are valid and string of found problems.
    """
    for hook in HOOKS:
        hook_path = get_hook_path(name=hook, path=project_context.path)
        if not hook_path.exists():
            message = WARNING + "Git hooks are not installed. " 'Use "renku githooks install" to install them. \n'
            return False, False, message

        with hook_path.open() as file_:
            actual_hook = _extract_renku_hook(file_)
        with StringIO(_read_resource(hook)) as file_:
            expected_hook = _extract_renku_hook(file_)

        if not expected_hook:
            message = WARNING + "Cannot check for existence of Git hooks.\n"
            return False, False, message

        if actual_hook != expected_hook:
            message = (
                WARNING + "Git hooks are outdated or not installed.\n"
                '  (use "renku githooks install --force" to update them) \n'
            )
            return False, False, message

    return True, False, None


def _extract_renku_hook(file):
    lines = [line.strip() for line in file if line.strip()]
    start = end = -1
    for index, line in enumerate(lines):
        if line.startswith("# RENKU HOOK."):
            start = index
        elif line.endswith("# END RENKU HOOK."):
            end = index
            break

    return lines[start:end] if 0 <= start <= end else []


def _read_resource(hook):
    return importlib_resources.files("renku.data").joinpath(f"{hook}.sh").read_bytes().decode("utf-8")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define repository checks for :program:`renku doctor`."""

from .activities import check_activity_dates, check_migrated_activity_ids
from .datasets import (
    check_dataset_files_outside_datadir,
    check_dataset_old_metadata_location,
    check_external_files,
    check_invalid_datasets_derivation,
    check_missing_files,
)
from .githooks import check_git_hooks_installed
from .migration import check_migration
from .project import check_project_id_group
from .storage import check_lfs_info
from .validate_shacl import check_datasets_structure, check_project_structure
from .workflow import check_activity_catalog, check_plan_id, check_plan_modification_date

# Checks will be executed in the order as they are listed in __all__. They are mostly used in ``doctor`` command to
# inspect broken things. The order of operations matters when fixing issues, so, don't sort this list.
__all__ = (
    "check_activity_catalog",
    "check_dataset_files_outside_datadir",
    "check_dataset_old_metadata_location",
    "check_datasets_structure",
    "check_git_hooks_installed",
    "check_invalid_datasets_derivation",
    "check_lfs_info",
    "check_migrated_activity_ids",
    "check_migration",
    "check_external_files",
    "check_missing_files",
    "check_project_id_group",
    "check_project_structure",
    "check_plan_modification_date",
    "check_plan_id",
    "check_activity_dates",
)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implement various context managers."""

import contextlib
import os
import sys
import time
from pathlib import Path
from typing import Dict, List, Optional, Union

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.interface.database_gateway import IDatabaseGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.ui.service.utils import normalize_git_url


@contextlib.contextmanager
def chdir(path):
    """Change the current working directory."""
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


class redirect_stdin(contextlib.ContextDecorator):
    """Implement missing redirect stdin based on ``contextlib.py``."""

    _stream = "stdin"

    def __init__(self, new_target):
        """Keep the original stream."""
        self._new_target = new_target
        # We use a list of old targets to make this CM re-entrant
        self._old_targets = []

    def __enter__(self):
        """Change the stream value."""
        self._old_targets.append(getattr(sys, self._stream))
        setattr(sys, self._stream, self._new_target)
        return self._new_target

    def __exit__(self, exception_type, exception_value, traceback):
        """Restore the stream value."""
        setattr(sys, self._stream, self._old_targets.pop())


def _wrap_path_or_stream(method, mode):
    """Open path with context or close stream at the end."""

    def decorator(path_or_stream):
        """Open the path if needed."""
        if isinstance(path_or_stream, (str, Path)):
            return method(Path(path_or_stream).open(mode))
        return method(path_or_stream)

    return decorator


class Isolation(contextlib.ExitStack):
    """Isolate execution."""

    CONTEXTS = {
        "stdin": _wrap_path_or_stream(redirect_stdin, "r"),
        "stdout": _wrap_path_or_stream(contextlib.redirect_stdout, "w"),
        "stderr": _wrap_path_or_stream(contextlib.redirect_stderr, "w"),
        "cwd": chdir,
    }

    def __init__(self, **kwargs):
        """Create a context manager."""
        super().__init__()
        for key, value in kwargs.items():
            if value is not None:
                self.enter_context(self.CONTEXTS[key](value))


@contextlib.contextmanager
def measure(message="TOTAL"):
    """Measure execution time of enclosing code block."""
    import time

    start = time.time()
    try:
        yield
    finally:
        end = time.time()
        total_seconds = float("%.2f" % (end - start))
        print(f"{message}: {total_seconds} seconds")


@contextlib.contextmanager
def renku_project_context(path, check_git_path=True):
    """Provide a project context with repo path injected."""
    from renku.core.util.git import get_git_path
    from renku.domain_model.project_context import project_context

    if check_git_path:
        path = get_git_path(path)

    path = normalize_git_url(str(path))

    with project_context.with_path(path=path):
        project_context.external_storage_requested = True
        yield project_context.path


@contextlib.contextmanager
@inject.autoparams("project_gateway", "database_gateway")
def with_project_metadata(
    project_gateway: IProjectGateway,
    database_gateway: IDatabaseGateway,
    read_only: bool = False,
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    description: Optional[str] = None,
    keywords: Optional[List[str]] = None,
    custom_metadata: Optional[Dict] = None,
):
    """Yield an editable metadata object.

    Args:
        project_gateway(IProjectGateway): Injected project gateway.
        database_gateway(IDatabaseGateway): Injected database gateway.
        read_only(bool): Whether to save changes or not (Default value = False).
        name(Optional[str]): Name of the project (when creating a new one) (Default value = None).
        namespace(Optional[str]): Namespace of the project (when creating a new one) (Default value = None).
        description(Optional[str]): Project description (when creating a new one) (Default value = None).
        keywords(Optional[List[str]]): Keywords for the project (when creating a new one) (Default value = None).
        custom_metadata(Optional[Dict]): Custom JSON-LD metadata (when creating a new project)
            (Default value = None).
    """
    from renku.domain_model.project import Project
    from renku.domain_model.project_context import project_context

    try:
        project = project_gateway.get_project()
    except ValueError:
        project = Project.from_project_context(
            project_context=project_context,
            name=name,
            namespace=namespace,
            description=description,
            keywords=keywords,
            custom_metadata=custom_metadata,
        )

    if read_only:
        yield project
    else:
        # NOTE: Set project so that ``project_context`` can be used inside the code
        project_gateway.update_project(project)

        yield project

        database_gateway.commit()


@contextlib.contextmanager
def Lock(filename: Union[Path, str], timeout: int = 0, mode: str = "shared", blocking: bool = False):
    """A file-based lock context manager."""
    import portalocker

    if mode == "shared":
        flags = portalocker.LOCK_SH
    elif mode == "exclusive":
        flags = portalocker.LOCK_EX
    else:
        raise errors.ParameterError(f"Mode can be 'shared' or 'exclusive' not '{mode}'")

    if not blocking:
        flags |= portalocker.LOCK_NB

    try:
        with portalocker.Lock(filename, timeout=timeout, flags=flags):
            yield
    except (portalocker.LockException, portalocker.AlreadyLocked) as e:
        raise errors.LockError(f"Cannot lock {e.__class__.__name__}")


@contextlib.contextmanager
def wait_for(delay: float):
    """Make sure that at least ``delay`` seconds are passed during the execution of the wrapped code block."""
    start = time.time()

    yield

    exec_time = time.time() - start
    if exec_time < delay:
        time.sleep(delay - exec_time)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version controller."""
import os

from renku import __version__
from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.serializers.version import VersionResponseRPC
from renku.ui.service.views import result_response


class VersionCtrl(ServiceCtrl):
    """Version controller."""

    RESPONSE_SERIALIZER = VersionResponseRPC()

    def to_response(self, minimum_version, maximum_version):
        """Serialize to service version response."""
        return result_response(
            VersionCtrl.RESPONSE_SERIALIZER,
            {
                "latest_version": __version__,
                "supported_project_version": SUPPORTED_PROJECT_VERSION,
                "cli_version": os.environ.get("RENKU_PROJECT_DEFAULT_CLI_VERSION") or __version__,
                "minimum_api_version": minimum_version.name,
                "maximum_api_version": maximum_version.name,
            },
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Models to represent a workflow file run templates."""

from __future__ import annotations

import hashlib
import uuid
from pathlib import Path
from typing import List, Optional, Union

from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.plan import Plan, validate_plan_name


class WorkflowFileCompositePlan(CompositePlan):
    """A workflow file composite plan."""

    plans: List[WorkflowFilePlan]

    def __init__(self, *, path: Union[Path, str], **kwargs):
        super().__init__(**kwargs)

        self.path: str = str(path)

    @staticmethod
    def generate_id(
        path: Optional[Union[Path, str]] = None, sequence: Optional[int] = None, uuid_only: bool = False, **_
    ) -> str:
        """Generate an identifier for Plan."""
        assert path, "Path is needed to generate id for WorkflowFileCompositePlan"

        # NOTE: Workflow file's root composite plan's ID is generated only based on the file's path. The ID might be
        # changed later if the plan is a derivative
        key = f"{path}" if sequence is None else f"{path}::{sequence}"
        key_bytes = key.encode("utf-8")

        uuid = hashlib.md5(key_bytes).hexdigest()[:32]  # nosec

        if uuid_only:
            return uuid
        return CompositePlan.generate_id(uuid=uuid)

    def assign_new_id(self, *, sequence: Optional[int] = None, **_) -> str:
        """Assign a new UUID or a deterministic."""
        new_id = (
            uuid.uuid4().hex
            if sequence is None
            else self.generate_id(path=self.path, sequence=sequence, uuid_only=True)
        )
        return super().assign_new_id(uuid=new_id)

    def is_equal_to(self, other: WorkflowFileCompositePlan) -> bool:
        """Return true if plan hasn't changed from the other plan."""
        return self.path == other.path and super().is_equal_to(other=other)


class WorkflowFilePlan(Plan):
    """Represent a Plan that is converted from a workflow file."""

    def __init__(self, *, path: Union[Path, str], **kwargs):
        super().__init__(**kwargs)

        self.path: str = str(path)

    @staticmethod
    def generate_id(
        path: Optional[Union[Path, str]] = None,
        name: Optional[str] = None,
        sequence: Optional[int] = None,
        uuid_only: bool = False,
        **_,
    ) -> str:
        """Generate an identifier for Plan."""
        assert path, "Path is needed to generate id for WorkflowFilePlan"
        assert name, "Name is needed to generate id for WorkflowFilePlan"

        key = f"{path}::{name}" if sequence is None else f"{path}::{name}::{sequence}"
        key_bytes = key.encode("utf-8")
        uuid = hashlib.md5(key_bytes).hexdigest()[:32]  # nosec
        if uuid_only:
            return uuid
        return Plan.generate_id(uuid=uuid)

    @staticmethod
    def validate_name(name: str):
        """Check a name for invalid characters."""
        validate_plan_name(name=name, extra_valid_characters="._-")

    @property
    def unqualified_name(self) -> str:
        """Name of the plan as appears in the workflow file definition."""
        return self.name.rsplit(".", maxsplit=1)[-1]

    def assign_new_id(self, *, sequence: Optional[int] = None, **_) -> str:
        """Assign a new UUID or a deterministic."""
        new_id = (
            uuid.uuid4().hex
            if sequence is None
            else self.generate_id(path=self.path, name=self.name, sequence=sequence, uuid_only=True)
        )
        return super().assign_new_id(uuid=new_id)

    def is_equal_to(self, other: WorkflowFilePlan) -> bool:
        """Return true if plan hasn't changed from the other plan."""
        return self.path == other.path and super().is_equal_to(other=other)

    def set_parameters_from_strings(self, params_strings: List[str]) -> None:
        """Set parameters by parsing parameters strings."""
        raise NotImplementedError



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General utility functions."""

import concurrent.futures
import os
import uuid
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import deal
from packaging.version import Version


def to_string(value: Optional[Any], strip: bool = False) -> str:
    """Return a string representation of value and return an empty string for None."""
    return "" if value is None else str(value).strip() if strip else str(value)


def to_semantic_version(value: str) -> Optional[Version]:
    """Convert value to SemVer."""
    try:
        return Version(value)
    except ValueError:
        return None


def is_uuid(value):
    """Check if value is UUID4.

    Copied from https://stackoverflow.com/questions/19989481/
    """
    try:
        uuid_obj = uuid.UUID(value, version=4)
    except ValueError:
        return False

    return str(uuid_obj) == value


@deal.pre(lambda _: _.rate > 0, message="Rate must be positive")
def parallel_execute(
    function: Callable[..., List[Any]], *data: Union[Tuple[Any, ...], List[Any]], rate: float = 1, **kwargs
) -> List[Any]:
    """Execute the function using multiple threads.

    Args:
        function(Callable[..., Any]): Function to parallelize. Must accept at least one parameter and returns a list.
        data(Union[Tuple[Any], List[Any]]): List of data where each of its elements is passed to a function's execution.
        rate(float): Number of executions per thread per second.

    Returns:
        List[Any]: A list of return results of all executions.

    """
    from renku.core.util import communication
    from renku.core.util.contexts import wait_for
    from renku.domain_model.project_context import project_context

    listeners = communication.get_listeners()

    def subscribe_communication_listeners(delay: float, path: Path, function, *data, **kwargs):
        try:
            for communicator in listeners:
                communication.subscribe(communicator)
            if not project_context.has_context(path):
                project_context.push_path(path)
            with wait_for(delay):
                return function(*data, **kwargs)
        finally:
            for communicator in listeners:
                communication.unsubscribe(communicator)

    # NOTE: Disable parallelization during tests for easier debugging
    if is_test_session_running():
        max_workers = 1
        delay = 0.0
    else:
        n_cpus = os.cpu_count() or 1
        max_workers = min(n_cpus + 4, 8)
        delay = max_workers / rate if len(data[0]) > max_workers else 0

    files = []
    with concurrent.futures.ThreadPoolExecutor(max_workers) as executor:
        futures = {
            executor.submit(subscribe_communication_listeners, delay, project_context.path, function, *d, **kwargs)
            for d in zip(*data)
        }

        for future in concurrent.futures.as_completed(futures):
            files.extend(future.result())

    return files


def is_test_session_running() -> bool:
    """Return if the code is being executed in a test and not called by user."""
    return "RENKU_RUNNING_UNDER_TEST" in os.environ



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of the project."""

from renku.command.command_builder import inject
from renku.command.util import WARNING
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util import communication
from renku.domain_model.project import Project
from renku.domain_model.project_context import project_context


@inject.autoparams("project_gateway")
def check_project_id_group(fix, project_gateway: IProjectGateway, **_):
    """Check that projects in groups have the correct id set.

    Args:
        fix: Whether to fix found issues.
        project_gateway: Injected project gateway.
        _: keyword arguments.

    Returns:
        Tuple of whether project id is valid, if an automated fix is available and string of found problems.
    """
    current_project = project_gateway.get_project()

    namespace, name = Project.get_namespace_and_name(
        remote=project_context.remote, repository=project_context.repository
    )

    if namespace is None or name is None:
        return True, False, None

    generated_id = Project.generate_id(namespace=namespace, name=name)

    if generated_id == current_project.id:
        return True, False, None

    if fix:
        communication.info(f"Fixing project id '{current_project.id}' -> '{generated_id}'")
        current_project.id = generated_id
        project_gateway.update_project(current_project)
        return True, False, None

    return (
        False,
        True,
        (
            WARNING
            + "Project id doesn't match id based on the current Git remote (use 'renku doctor --fix' to fix it):"
            f"\n\t'{current_project.id}' -> '{generated_id}'"
        ),
    )
